from typing import (
    Union,
    Dict,
    Iterator,
    List,
    Optional,
)

from pathlib import Path

from ._common_imports import (
    classad,
    handle_t,
)

from collections.abc import MutableMapping

from .htcondor2_impl import (
    _submit_init,
    _submit__getitem__,
    _submit__setitem__,
    _submit_keys,
    _submit_expand,
    _submit_getqargs,
    _submit_setqargs,
    _submit_from_dag,
    _display_dag_options,
    _submit_set_submit_method,
    _submit_get_submit_method,
    _submit_issue_credentials,
    _submit_itemdata,
    HTCondorException,
)

from ._submit_method import SubmitMethod

DefaultItemData = object()

#
# MutableMapping provides generic implementations for all mutable mapping
# methods except for __[get|set|del]item__(), __iter__, and __len__.
#

class Submit(MutableMapping):

    # This is exceedingly clumsy, but matches the semantics from version 1.
    def __init__(self,
        input : Union[Dict[str, str], str] = None,
        queue : str = None,
        ** kwargs
    ):
        '''
        An object representing a job submit description.  It uses the same
        language as *condor_submit.*  Values in the submit description language
        have no type; they are all strings.

        If `input` is a dictionary, it will be serialized into the submit
        language.  Otherwise, `input` is a string in the submit language.
        This string may end with a queue statement.  You may not specify a
        queue statement in an `input` dictionary; use the `queue` keyword
        argument to specify its arguments, instead.  (You may preface the
        the keyword's string value with "queue ", if you prefer.)

        If other keyword arguments are supplied, they will update this object
        after the submit-language string, if any, is parsed.

        This object implements the Python dictionary protocol, except
        that attempts to ``del()`` entries will raise a
        :class:`NotImplementedError`
        because the underlying logic does not distinguish between a
        missing key and a key whose value is the empty string.  (This seems
        better than allowing you to ``del(s[key])`` but having ``key`` still
        appear in ``s.keys()``.)
        '''
        self._handle = handle_t()

        # if isinstance(input, dict) and len(input) != 0:
        if isinstance(input, dict):
            pairs = []
            for key,value in input.items():
                if not isinstance(key, str):
                    raise TypeError("key must be a string")
                if key.casefold() == "queue".casefold():
                    raise ValueError("the queue statement can not be specified in a dictionary")
                # This was an undocumented feature in version 1, and it's
                # likely to be useless.  This implementation assumes that
                # str(classad.ExprTree) is always valid in a submit-file.
                if isinstance(value, classad.ClassAd):
                    pairs.append(f"{key} = {repr(value)}")
                else:
                    # In version 1, we (implicitly) document the dict as
                    # being from strings to strings, but don't enforce it;
                    # the code basically just calls str() on the value and
                    # hopes for the best.  This is probably a mistake in
                    # general, but it's really convenient, and Ornithology
                    # depends on being able to submit bools and ints as
                    # those types.  Luckily, those are easy to check for
                    # specifically and unambiguous to handle.
                    if isinstance(value, int):
                        value = str(value)
                    elif isinstance(value, float):
                        value = str(value)
                    elif isinstance(value, bool):
                        value = str(value)
                    # This is cheating, but we use it _a lot_ in the test suite.
                    elif isinstance(value, Path):
                        value = str(value)
                    elif not isinstance(value, str):
                        raise TypeError("value must be a string")
                    pairs.append(f"{key} = {value}")
            s = "\n".join(pairs)
            _submit_init(self, self._handle, s)
        elif isinstance(input, str):
            _submit_init(self, self._handle, input)
        # elif input is None or isinstance(input, dict) and len(input) == 0:
        elif input is None:
            _submit_init(self, self._handle, None)
        else:
            raise TypeError("input must be a dictionary mapping string to strings, a string, or None")

        self.update(kwargs)

        if queue is not None:
            if isinstance(queue, str):
                self.setQArgs(queue)
            else:
                raise TypeError("queue must be a string")


    def __getitem__(self, key):
        if not isinstance(key, str):
            raise TypeError("key must be a string")
        return _submit__getitem__(self, self._handle, key)


    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise TypeError("key must be a string")
        if not isinstance(value, str):
            raise TypeError("value must be a string")
        return _submit__setitem__(self, self._handle, key, value)


    def __delitem__(self, key):
        # We can actually fake this (using NULL values), but should we?
        raise NotImplementedError("Submit object keys can not be removed, but the empty string is defined to have no effect.")


    #
    # Python's documentation is confusing and somewhat self-contradictory,
    # but the build-in `dict` tries really hard to raise a `RuntimeError`
    # if the dictionary is modified while iterating.
    #
    # This suggests that it's OK for __iter__ to get the list of keys
    # and then `yield` them one-at-a-time.  This is good, since the C++
    # `SubmitHash` almost certainly can't handle anything else.
    #
    # In version 1, keys(), items(), and values() all explicitly ignored
    # the submit hash's default values, but __getitem__() did not.  This
    # was madness.  Implementing htcondor2.Submit as a MutableMapping
    # subclass eliminates the possibility of those functions drifting out
    # synch with __iter__(); the current implementation of __len__() is
    # slow, but should also always agree with whatever we do in __iter__().
    #
    # It would be possible to change __getitem__() to ignore the default
    # values as well (in C++, or less efficiently in Python), but that
    # doesn't seem helpful.
    #
    # Ignoring default values actually turns out to be necessary to correctly
    # (re)generate submit files from submit hashes.
    #
    def __iter__(self):
        keys = _submit_keys(self, self._handle)
        if keys is not None:
            for key in keys.split('\0'):
                yield(key)


    def __len__(self):
        keys = _submit_keys(self, self._handle)
        if keys is None:
            return 0
        return len(keys.split('\0'))


    def __str__(self):
        rv = ''
        for key, value in self.items():
            rv = rv + f"{key} = {value}\n"
        qargs = _submit_getqargs(self, self._handle)
        if qargs is not None:
            if qargs == '':
                rv = rv + "queue\n"
            else:
                rv = rv + "queue " + qargs + "\n"
        return rv


    def __repr__(self):
        return f'Submit({self})'


    def expand(self, attr : str) -> str:
        '''
        Expand all macros for the given attribute.

        :param str attr: The name of the attribute to expand.
        :return: The value of the given attribute, with all macros expanded.
        :rtype: str
        '''
        if not isinstance(attr, str):
            raise TypeError("attr must be a string")
        return _submit_expand(self, self._handle, attr)


    # In version 1, the documentation widly disagrees with the implementation;
    # this is what the implementation actually required.
    def jobs(self,
        count : int = 0,
        itemdata : iter = None,
        clusterid : int = 1,
        procid : int = 0,
        qdate : int = 0,
        owner : str = "",
    ):
        """
        This function is not currently implemented.

        :param count:
        :param itemdata:
        :param clusterid:
        :param procid:
        :param qdate:
        :param owner:
        """
        raise NotImplementedError("Let us know what you need this for.")


    # In version 1, the documentation widly disagrees with the implementation;
    # this is what the implementation actually required.
    def procs(self,
        count : int = 0,
        itemdata : iter = None,
        clusterid : int = 1,
        procid : int = 0,
        qdate : int = 0,
        owner : str = "",
    ):
        """
        This function is not currently implemented.

        :param count:
        :param itemdata:
        :param clusterid:
        :param procid:
        :param qdate:
        :param owner:
        """
        raise NotImplementedError("Let us know what you need this for.")


    def getQArgs(self) -> str:
        """
        Return the arguments to the queue statement.
        """
        return _submit_getqargs(self, self._handle)


    def setQArgs(self, args : str):
        '''
        Set the arguments to the queue statement.  These arguments replace
        the arguments, if any, passed to the original constructor.

        :param args:  The arguments.  May start with "queue ".
        '''
        if not isinstance(args, str):
            raise TypeError("args must be a string")

        # We can always add a ValueError here if it's requested.
        if args.casefold().startswith("queue ".casefold()):
            args = args[6:]

        _submit_setqargs(self, self._handle, args)


    def itemdata(self, qargs : Optional[str] = None) -> Union[ Iterator[str], Iterator[dict] ]:
        '''
        Returns an iterator over the itemdata specified by the given
        arguments to a queue statement,
        suitable for passing to :meth:`schedd.Submit`.

        ``s.itemdata()`` is equivalent to ``s.itemdata(s.getQAargs())``.

        :param qargs:  A set of arguments for a queue statement.  May
                       start with "queue ".
        '''
        s = self
        if qargs is not None:
            # It's legal for the qargs to reference submit macros,
            # so we have to build a duplicate of self.  We could do
            # this one the C++ side, instead, but we'd have to figure
            # out copying the SubmitBlob, instead.  (Or verify that
            # temporarily setting and resetting qargs doesn't change
            # anything, which seems tedious and unlikely.)
            s = Submit(str(self))
            s.setQArgs(qargs)
        (keys_str, values_str) = _submit_itemdata(s, s._handle)

        if values_str is None:
            return None
        elif keys_str is None:
            return iter(values_str.split("\n"))
        else:
            keys = keys_str.split("\n")
            values = values_str.split("\n")
            rv = []
            for value in values:
                if not value:
                    raise ValueError("invalid qargs")
                d = {}
                if "\x1F" in value:
                    v = value.split("\x1F")
                else:
                    v = [value]
                for i in range(0, len(keys)):
                    d[keys[i]] = v[i]
                rv.append(d)
            return iter(rv)


    def setSubmitMethod(self,
        method_value : Union[SubmitMethod, int] = -1,
        allow_reserved_values : bool = False
    ):
        """
        By default, jobs created by this class record (in the job ad
        attribute :ad-attr:`JobSubmitMethod`) that they were submitted via
        the  Python bindings.  Calling this method before submitting allows
        you to change that.  Values between ``0`` (inclusive) and ``100``
        (exclusive) are reserved for :class:`htcondor2.SubmitMethod`
        and can't be set unless the ``allowed_reserved_values`` flag is
        :const:`True`.  Values less than `0` may be used to remove
        ``JobSubmitMethod`` from the job ad entirely.

        :param method_value:  The method to record.
        :param allowed_reserved_values:  Set to true to used a reserved ``method_value``.
        """
        if isinstance(method_value, SubmitMethod):
            method_value = int(method_value)
        if 0 <= method_value and method_value < 100:
            if not allow_reserved_values:
                # This was HTCondorValueError in version 1.
                raise ValueError("Submit method value must be 100 or greater, or allowed_reserved_values must be True.")
        _submit_set_submit_method(self._handle, method_value)


    def getSubmitMethod(self) -> int:
        """
        Returns the submit method.

        :return:  The integer value of the submit method.  The symbolic
                  value can be obtained from the :class:`SubmitMethod`
                  enumeration.
        """
        return _submit_get_submit_method(self._handle)


    @staticmethod
    def from_dag(filename : str, options : Dict[str, Union[int, bool, str]] = {}) -> "Submit":
        """
        Returns a :class:`Submit` object that can be used to submit the
        DAG specified in the file `filename`.

        :param filename:  The DAG description file.
        :param options:  A dictionary of *condor_submit_dag* command-line
            options and their values.  See :meth:`from_dag_options`.
        """
        if not isinstance(options, dict):
            raise TypeError("options must be a dict")
        if not Path(filename).exists():
            raise HTCondorException(f"{filename} does not exist")

        # Convert from version 1 names to the proper internal names.
        internal_options = {}
        for key, value in options.items():
            if not isinstance(key, str):
                raise TypeError("options keys must by strings")
            elif len(key) == 0:
                raise ValueError("options key is empty")
            internal_key = _NewOptionNames.get(key.lower(), key)
            if not isinstance(value, (int, bool, str)):
                raise TypeError("options values must be int, bool, or str")
            internal_options[internal_key] = str(value)

        subfile = _submit_from_dag(filename, internal_options)
        subfile_text = Path(subfile).read_text()
        return Submit(subfile_text)


    def issue_credentials(self) -> Union[str, None]:
        '''
        Issue credentials for this job description.

        .. note::
            As with :tool:`condor_submit`, this assumes that the local
            machine is the target AP.

        :return:  A string containing a URL that the submitter must visit
                  in order to complete an OAuth2 flow, or :py:obj:`None`
                  if no such visit is necessary.
        '''
        return _submit_issue_credentials(self._handle)


    @staticmethod
    def from_dag_options():
        """
        Prints a table of valid keys for the ``options`` dictionary
        passed to :meth:`from_dag`, the type of each key, and the
        short description from *condor_submit_dag*'s ``-help`` output.

        This function is useful if you don't have an installed copy of
        :tool:`condor_submit_dag` or if this module and that command-line
        tool could be from different versions.

        .. note::

            Not all options available to *condor_submit_dag* exist for
            the python bindings.
        """
        _display_dag_options()


# List does not include options which vary only in capitalization.
_NewOptionNames = {
    "dagman":                   "DagmanPath",
    "schedd-daemon-ad-file":    "ScheddDaemonAdFile",
    "schedd-address-file":      "ScheddAddressFile",
    "alwaysrunpost":            "PostRun",
    "debug":                    "DebugLevel",
    "outfile_dir":              "OutfileDir",
    "config":                   "ConfigFile",
    "batch-name":               "BatchName",
    "load_save":                "SaveFile",
    "do_recurse":               "Recurse",
    "update_submit":            "UpdateSubmit",
    "import_env":               "ImportEnv",
    "include_env":              "GetFromEnv",
    "insert_env":               "AddToEnv",
    "valgrind":                 "RunValgrind",
    "suppress_notification":    "SuppressNotification",
    "dorecov":                  "DoRecovery",
}



from typing import List, Optional

from ._common_imports import (
    classad,
    Collector,
    DaemonType,
)

from .htcondor2_impl import (
    _negotiator_command,
    _negotiator_command_return,
    _negotiator_command_user,
    _negotiator_command_user_return,
    _negotiator_command_user_value,
    _dprintf_dfulldebug,
)


def _validate_user(user):
    if not isinstance(user, str):
        raise TypeError("user must be a string")
    if '@' not in user:
        # This was HTCondorValueError in version 1.
        raise ValueError("You must specify the submitter (user@uid.domain)")


class Negotiator():
    """
    The negotiator client.  Query and manage resource usage.

    Consult `User Priorities and Negotiation <https://htcondor.readthedocs.io/en/latest/admin-manual/cm-configuration.html#configuration-for-central-managers>`_ before using these functions.
    """

    def __init__(self, location : Optional[classad.ClassAd] = None):
        """
        :param location:  A ClassAd with a ``MyAddress`` attribute, such as
            might be returned by :meth:`htcondor2.Collector.locate`.  :py:obj:`None` means the
            default pool negotiator.
        """
        if location is None:
            c = Collector()
            location = c.locate(DaemonType.Negotiator)

        if not isinstance(location, classad.ClassAd):
            raise TypeError("location must be a ClassAd")

        self._addr = location['MyAddress']
        # We never actually use this for anything.
        # self._version = location['CondorVersion']


    def deleteUser(self, user : str) -> None:
        """
        Delete all records of an accounting principal from the negotiator’s
        accounting.

        :param user:  A fully-qualifed (``user@domain``) accounting principal.
        """
        _validate_user(user)
        _negotiator_command_user(self._addr, self._DELETE_USER, user)


    def getPriorities(self, rollup : bool = False) -> List[classad.ClassAd]:
        """
        Retrieve the pool accounting information as a list of
        `accounting ClassAds <http://htcondor.readthedocs.io/en/latest/classad-attributes/accounting-classad-attributes.html>`_.

        :param rollup:  It :py:obj:`True`, the accounting information
            that applies to heirarchical group quotas will be summed for
            groups and subgroups.
        """
        command = self._GET_PRIORITY
        if rollup:
            command = self._GET_PRIORITY_ROLLUP
        returnAd = _negotiator_command_return(self._addr, command)
        return _convert_numbered_attributes_to_list_of_ads(returnAd)


    def getResourceUsage(self, user : str) -> List[classad.ClassAd]:
        """
        Retrieve the resources (slots) assigned to the specificied
        accounting principal as a list of ClassAds.  The names of the
        attributes and their types are currently undocumented.

        :param user:  A fully-qualifed (``user@domain``) accounting principal.
        """
        _validate_user(user)
        returnAd = _negotiator_command_user_return(self._addr, self._GET_RESLIST, user)
        return _convert_numbered_attributes_to_list_of_ads(returnAd)


    def resetAllUsage(self):
        """
        Set the accumulated usage of all accounting principals to zero.
        """
        _negotiator_command(self._addr, self._RESET_ALL_USAGE)


    def resetUsage(self, user : str) -> None:
        """
        Set the accumulated usage of the specified accounting principal to zero.

        :param user:  A fully-qualifed (``user@domain``) accounting principal.
        """
        _validate_user(user)
        _negotiator_command_user(self._addr, self._RESET_USAGE, user)


    def setBeginUsage(self, user : str, when : int) -> None:
        """
        Set the beginning of the specified accounting principal's usage.

        :param user:  A fully-qualifed (``user@domain``) accounting principal.
        :param when:  A Unix timestamp.
        """
        _validate_user(user)
        if isinstance(when, float):
            _dprintf_dfulldebug("setBeginUsage(): `when` specified as float, ignoring fractional part\n")
        _negotiator_command_user_value(self._addr, self._SET_BEGINTIME, user, int(when))


    # This incorrectly -- and brokenly -- takes a float in version 1.
    def setCeiling(self, user : str, ceiling : int) -> None:
        """
        Set the submitter ceiling for the specific accounting principal.

        :param user:  A fully-qualifed (``user@domain``) accounting principal.
        :param ceiling:  A value greater than or equal to ``-1``.
        """
        _validate_user(user)
        if isinstance(ceiling, float):
            _dprintf_dfulldebug("setCeiling(): `ceiling` specified as float, ignoring fractional part\n")
        if ceiling <= -1:
            # This was HTCondorValueError in version 1.
            raise ValueError("Ceiling must be greater than -1.")
        _negotiator_command_user_value(self._addr, self._SET_CEILING, user, int(ceiling))


    def setLastUsage(self, user : str, when : int) -> None:
        """
        Set the end of the specified accounting principal's usage.

        :param user:  A fully-qualifed (``user@domain``) accounting principal.
        :param when:  A Unix timestamp.
        """
        _validate_user(user)
        _negotiator_command_user_value(self._addr, self._SET_LASTTIME, user, int(when))


    # This works exactly the same way it did version 1, including the ugly
    # error message from not waiting for the reply.
    def setFactor(self, user : str, factor : float) -> None:
        """
        Set the priority factor of the specified accounting principal.

        :param user:  A fully-qualifed (``user@domain``) accounting principal.
        :param factor:  A value greater than or equal to ``1.0``.
        """
        _validate_user(user)
        if factor < 1:
            raise ValueError("Priority factors must be >= 1")
        _negotiator_command_user_value(self._addr, self._SET_PRIORITYFACTOR, user, float(factor))


    def setPriority(self, user : str, priority : float) -> None:
        """
        Set the priority of the specified accounting principal.

        :param user:  A fully-qualifed (``user@domain``) accounting principal.
        :param priority:  A value greater than or equal to ``0.0``.
        """
        _validate_user(user)
        if priority < 0:
            raise ValueError("User priority must be non-negative")
        _negotiator_command_user_value(self._addr, self._SET_PRIORITY, user, float(priority))


    def setUsage(self, user : str, usage : float) -> None:
        """
        Set the usage of the specified accounting principal.

        :param user:  A fully-qualifed (``user@domain``) accounting principal.
        :param usage:  The usage in hours.
        """
        _validate_user(user)
        if usage < 0:
            raise ValueError("Usage must be non-negative")
        _negotiator_command_user_value(self._addr, self._SET_ACCUMUSAGE, user, float(usage))


    _SCHED_VERS          = 400
    _SET_PRIORITY        = _SCHED_VERS + 49
    _GET_PRIORITY        = _SCHED_VERS + 51
    _RESET_USAGE         = _SCHED_VERS + 58
    _SET_PRIORITYFACTOR  = _SCHED_VERS + 59
    _RESET_ALL_USAGE     = _SCHED_VERS + 60
    _GET_RESLIST         = _SCHED_VERS + 63
    _DELETE_USER         = _SCHED_VERS + 82
    _SET_ACCUMUSAGE      = _SCHED_VERS + 94
    _SET_BEGINTIME       = _SCHED_VERS + 95
    _SET_LASTTIME        = _SCHED_VERS + 96
    _GET_PRIORITY_ROLLUP = _SCHED_VERS + 114
    _SET_CEILING         = _SCHED_VERS + 125
    _SET_FLOOR           = _SCHED_VERS + 130


def _convert_numbered_attributes_to_list_of_ads(ad : classad.ClassAd):
    attrs = []
    for attr in ad.keys():
        if attr.endswith("1"):
            attrs.append(attr[:-1])

    l = []
    index = 1
    carry_on = True
    while carry_on:
        carry_on = False

        c = classad.ClassAd()
        for attr in attrs:
            indexedAttr = f"{attr}{index}"
            e = ad.get(indexedAttr)
            if e is not None:
                carry_on = True
                c[attr] = e

        if carry_on:
            l.append(c)

        index += 1

    return l



import os
import sys

from docutils import nodes, utils
from docutils.parsers.rst import Directive
from sphinx import addnodes
from sphinx.errors import SphinxError
from sphinx.util.nodes import split_explicit_title, process_index_entry, set_role_source_info
from htc_helpers import *

FUNCTION_DEF_COUNT = {}

def dump(obj):
    for attr in dir(obj):
        print("obj.%s = %r" % (attr, getattr(obj, attr)))

def classad_func_def_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    global FUNCTION_DEF_COUNT
    docname = inliner.document.settings.env.docname

    function, info = custom_ext_parser(text, "<", ">")

    if ")" not in function or "(" not in function or " " not in function[:function.find("(")]:
        # Expect 'ReturnType FunctionName([args])'
        warn(f"{docname}:{lineno} | Invalid ClassAd Function '{function}' expects 'Return function(...)'")
        return [make_inline_literal_node(text)], []

    fn_end = function.find("(")
    fn_begin = function[:fn_end].find(" ") + 1
    func_name = function[fn_begin : fn_end]

    if func_name in FUNCTION_DEF_COUNT:
        FUNCTION_DEF_COUNT[func_name] += 1
        span_id = f"{func_name}({FUNCTION_DEF_COUNT[func_name]})"
        headerlink_node = make_headerlink_node(span_id, options)
    else:
        span_id = f"{func_name}()"
        headerlink_node = make_headerlink_node(span_id, options)
        FUNCTION_DEF_COUNT[func_name] = 0

    inline_node = make_inline_literal_node(function, span_id)

    index_node = addnodes.index()
    index_node['entries'] = process_index_entry(f"pair: {func_name}() ; ClassAd Functions", f"{func_name}()")
    set_role_source_info(inliner, lineno, index_node)

    return [index_node, inline_node, headerlink_node], []

def setup(app):
    app.add_role("classad-function-def", classad_func_def_role)




import htcondor2
import classad2
from htcondor2._utils.ansi import underline
from pathlib import Path

from htcondor_cli.convert_ad import echo_ads_to_daemon_status, health_str
from htcondor_cli.noun import Noun
from htcondor_cli.verb import Verb

class Status(Verb):
    """
    Displays status information all/specified access points
    """

    options = {
        "hostnames": {
            "args": ("hostname",),
            "nargs": "*",
            "action": "append",
            "help": "Specific Access Point hostnames to query",
        }
    }

    def __init__(self, logger, **options):
        hostnames = [host for sublist in options["hostname"] for host in sublist]
        constraint = None
        if len(hostnames) > 0:
            constraint = " || ".join(f'Machine == "{host}"' for host in hostnames)

        try:
            collector = htcondor2.Collector()
            ads = collector.query(ad_type=htcondor2.AdType.Schedd, constraint=constraint)
        except Exception:
            ads = []

        if len(ads) == 0:
            if len(hostnames):
                raise RuntimeError("Failed to locate specified Access Point(s)")
            else:
                local_schedd_ad = Path(htcondor2.param["SCHEDD_DAEMON_AD_FILE"])
                if local_schedd_ad.exists():
                    ads.append(classad2.parseOne(local_schedd_ad.open().read()))
                else:
                    raise RuntimeError("Failed to locate any Access Point information")

        ap_status = {}
        width = 0

        for daemon, status, ad in echo_ads_to_daemon_status(ads):
            assert daemon == "SCHEDD"
            host = status["Machine"]
            width = len(host) if len(host) > width else width

            if host in hostnames:
                hostnames.remove(host)

            if host in ap_status:
                ap_status[host]["HP"] += status["HealthPoints"]
                ap_status[host]["Count"] += 1
            else:
                ap_status[host] = {"HP": status["HealthPoints"], "Count": 1, "RDCDC" : ad["RecentDaemonCoreDutyCycle"]}

        if len(ap_status) > 0:
            columns = ["Access Point", "Health", "DutyCycle"]
            logger.info(underline("{0: ^{width}}   {1}   {2}".format(*columns, width=width)))
            for host, status in ap_status.items():
                rdcdc = status["RDCDC"] * 100
                health = health_str(status['HP'], status['Count'])
                logger.info(f"{host: <{width}}   {health: >6}   {rdcdc: >10.2}%")

        if len(hostnames) > 0:
            print(
                "{0}Failed to find the following hosts:".format(
                    "" if len(ap_status) == 0 else "\n"
                )
            )
            for host in hostnames:
                print(host)

        logger.info("")


class AccessPoint(Noun):
    """
    Run operations on pool access points
    """

    class status(Status):
        pass

    @classmethod
    def verbs(cls):
        return [cls.status]



#!/usr/bin/env pytest

import os

from htcondor2 import Schedd
from classad2 import ClassAd

from pytest_httpserver import HTTPServer
from werkzeug.wrappers import Response

from ornithology import (
    action,
    Condor,
    ClusterState,
)


# Unset HTTP_PROXY for correct operation in Docker containers
lowered = dict()
for k in os.environ:
    lowered[k.lower()] = k
os.environ.pop(lowered.get("http_proxy", "http_proxy"), None)

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@action
def the_condor(test_dir):
    local_dir = test_dir / "condor"

    with Condor(
        local_dir=local_dir,
        config={
            "STARTD_ENVIRONMENT":           ";http_proxy=;https_proxy=",
            "SHADOW_DEBUG":                 "D_TEST D_CATEGORY",
            "CHECK_INPUT_FILE_TRANSFER":    "TRUE",
        },
    ) as the_condor:
        yield the_condor

@action
def the_server():
    with HTTPServer() as server:
        yield server


def the_response_function():
    for word in ['the', 'sizeless', 'response']:
        yield word


@action
def the_URLs(the_server):
    the_server.expect_request("/absent").respond_with_data(status=404)
    the_server.expect_request("/present").respond_with_data("Great success!")

    # You MUST invoke a generator here; the documenation lies when it says
    # that any iterable will "switch into streaming mode", which for our
    # purposes just means "will stop generating a Content-Length header".
    r = Response(
        response=the_response_function(),
    )
    the_server.expect_request("/sizeless").respond_with_response(r)

    return (
        f"http://localhost:{the_server.port}/absent",
        f"http://localhost:{the_server.port}/present",
        f"http://localhost:{the_server.port}/sizeless",
    )


# This function is rather magical.
@action
def the_expected_results(test_dir, the_URLs):
    (absentURL, presentURL, presentSizelessURL) = the_URLs
    input_text = test_dir / "input.text"

    the_list = [
        [input_text.as_posix(), "true", "0", "true"],
        [absentURL,             "false", None, "false"],
        [presentURL,            "true", "14", "true"],
        [presentSizelessURL,    "true", None, "false"],
    ]

    return { entry[0]: entry for entry in the_list }


@action
def the_held_job(the_condor, test_dir, path_to_sleep, the_URLs):
    (absentURL, presentURL, presentSizelessURL) = the_URLs
    input_text = test_dir / "input.text"
    input_text.touch()

    the_description = {
        "universe":                 "vanilla",
        "executable":               path_to_sleep,
        "arguments":                1,
        "transfer_executable":      False,
        "should_transfer_files":    True,
        "log":                      test_dir / "sleep.log",
        "request_cpus":             1,
        "request_memory":           1,
        "transfer_input_files":     ",".join([
            input_text.as_posix(),
            # We can't presently reliably test osdf or https.
            absentURL,
            presentURL,
            presentSizelessURL,
            # These are not verified but should be counted in TransferInputFileCounts
            "null://bar",
            "null://foo",
            "debug://success",
        ]),
    }
    job_handle = the_condor.submit(the_description, count=1)

    assert job_handle.wait(
        condition=ClusterState.all_held,
        fail_condition=ClusterState.any_complete,
        timeout=60,
    )

    return job_handle


@action
def the_job_ad(the_condor, the_held_job):
    ads = None
    with the_condor.use_config():
        ads = Schedd().query(
            constraint=f"ClusterId=={the_held_job.clusterid}",
            projection=["TransferInputFileCounts"]
        )
    return ads


@action
def the_expected_count_ad():
    return ClassAd({
        "CEDAR": 1,
        "HTTP": 3,
        "NULL": 2,
        "DEBUG": 1,
    })


class TestInputFileCheck:

    def test_check_table(self, the_held_job, the_condor, the_expected_results, the_URLs):
        the_shadow_log = the_condor.shadow_log.open()
        for line in the_shadow_log.read():
            if "D_TEST" in line.tags:
                if "checkInputFileTransfer():" in line.line:
                    # Remove the prefix from the list.
                    actual = line.line.split("\t")[1:]
                    expected = the_expected_results[actual[0]]

                    # If the size isn't known, change the reported size so
                    # this code doesn't have to know what (size_t)-1 is.
                    if expected[3] != "true":
                        actual[2] = None

                    logger.debug(f"  actual = {actual}")
                    logger.debug(f"expected = {expected}")

                    assert actual == expected

    def test_check_protocol_counts(self, the_job_ad, the_expected_count_ad):
        assert the_job_ad is not None
        assert len(the_job_ad) == 1
        ad = the_job_ad[0]
        assert "TransferInputFileCounts" in ad
        assert the_expected_count_ad == ad["TransferInputFileCounts"]



from .classad2_impl import _handle as handle_t

from .classad2_impl import (
    _classad_init,
    _classad_init_from_string,
    _classad_init_from_dict,
    _classad_to_string,
    _classad_to_repr,
    _classad_get_item,
    _classad_set_item,
    _classad_del_item,
    _exprtree_eq,
    _classad_size,
    _classad_keys,
    _classad_parse_next,
    _classad_parse_next_fd,
    _classad_quote,
    _classad_unquote,
    _classad_flatten,
    _classad_internal_refs,
    _classad_external_refs,
    _classad_print_json,
    _classad_print_old,
    _classad_last_error,
    _classad_version,
    ClassAdException,
)

import io

# So that the typehints match version 1.
from ._parser_type import ParserType as Parser

from ._value import Value

from collections.abc import MutableMapping
from datetime import datetime, timedelta, timezone

from typing import (
    Iterator,
    Union,
    IO,
    Optional,
    List,
)

#
# MutableMapping provides generic implementations for all mutable mapping
# methods except for __[get|set|del]item__(), __iter__, and __len__.
#

class ClassAd(MutableMapping):
    R"""
    A :class:`ClassAd` is a :class:`collections.abc.MutableMapping` with
    class:`str`\ing keys that represents a "record expression" in the
    `ClassAd language <https://htcondor.org/classad/refman/>`_, more
    usually called "a ClassAd".

    Values in a ClassAd have types, and those types are represented as
    built-in Python types --

      * boolean (:class:`bool`)
      * integer (:class:`int`)
      * float (:class:`float`)
      * string (:class:`str`)
      * absolute time value (:class:`datetime.datetime`)
      * relative time value (:class:`float`)
      * list (:class:`list`)

    -- with the following exceptions:

      * ClassAd (:class:`ClassAd`)
      * undefined (:data:`classad2.Value.Undefined`)
      * error (:data:`classad2.Value.Error`)
      * expression (:class:`ExprTree`)

    When setting a value, :py:obj:`None` is converted to
    :data:`classad2.Value.Undefined` and :class:`dict`\s
    are converted to :class:`ClassAd`\s.

    Expressions are always evaluated lazily, so setting a value to
    :class:`ExprTree` ``2 + 2`` will not result in the value being
    the :class:`int` ``4``; this also applies when constructing or
    parsing :class:`ClassAd`\s or :class:`ExprTree`\s.
    """

    def __init__(self, input : Optional[Union[str, dict]] = None):
        """
        :param input:
            * If :py:obj:`None`, create an empty ClassAd.
            * If a :class:`str`, parse the :const:`ParserType.New` format
              string and create the corresponding ClassAd.
            * If a :class:`dict`, convert the dictionary to a ClassAd.
        """
        self._handle = handle_t()

        if input is None:
            _classad_init(self, self._handle)
            return
        if isinstance(input, dict):
            _classad_init_from_dict(self, self._handle, input)
            return
        if isinstance(input, str):
            _classad_init_from_string(self, self._handle, input)
            return
        raise TypeError("input must be None, str, or dict")


    def __repr__(self):
        return _classad_to_repr(self._handle)


    def __str__(self):
        return _classad_to_string(self._handle)


    # This tells the pickle module to reconstruct ClassAd objects by
    # calling the constructor (that is, `__init__()`) with the given
    # tuple of arguments, intead of its default method, which tries
    # to pickle the handle, which is never going to work.
    def __reduce__(self):
        return (ClassAd, (repr(self),))


    def eval(self, attr : str):
        """
        Evaluate the attribute named by *attr* in the context of this ClassAd.

        :param attr: The attribute to evaluate.
        """
        expression = self.lookup(attr)
        return expression.eval(self)


    def externalRefs(self, expr: "ExprTree") -> List[str]:
        """
        Returns the list of attributes referenced by the expression
        *expr* which are **not** defined by this ClassAd.

        :param expr: The expression.
        """
        if not isinstance(expr, ExprTree):
            raise TypeError("expr must be an ExprTree")

        string_list = _classad_external_refs(self._handle, expr._handle)
        return string_list.split(",")


    def flatten(self, expr : "ExprTree") -> "ExprTree":
        """
        Partially evaluate the expression in the context of this ad.
        Attributes referenced by the expression which are not defined
        by this ClassAd are not evaluated, but other attributes and
        subexpressions are.  Thus, ``RequestMemory * 1024 * 1024`` will
        evaluate to the amount of requested memory in bytes if this
        ClassAd defines the ``RequestMemory`` attribute, and to
        ``RequestMemory * 1048576`` otherwise.

        :param expr: The expression.
        """
        if not isinstance(expr, ExprTree):
            raise TypeError("expr must be an ExprTree")
        return _classad_flatten(self._handle, expr._handle)


    def internalRefs(self, expr: "ExprTree") -> List[str]:
        """
        Returns the list of attributes referenced by the expression *expr*
        which **are** defined by this ClassAd.

        :param expr: The expression.
        """
        if not isinstance(expr, ExprTree):
            raise TypeError("expr must be an ExprTree")

        string_list = _classad_internal_refs(self._handle, expr._handle)
        return string_list.split(",")


    def lookup(self, attr : str) -> "ExprTree":
        """
        Return the :class:`ExprTree` named by *attr*.

        :param attr: The attribute to look up.
        """
        if not isinstance(attr, str):
            raise TypeError("ClassAd keys are strings")
        return _classad_get_item(self._handle, attr, False)


    def matches(self, ad : "ClassAd") -> bool:
        """
        Evaluate the ``requirements`` attribute of the given *ad* in the
        context of this one, converting :class:`int` and :class:`float`
        results to :class:`bool`.

        :param ad: The ClassAd to test for matching.
        :return:  The evaluated or converted boolean, or ``False``.
        """
        if not isinstance(ad, ClassAd):
            raise TypeError("ad must be a ClassAd")

        try:
            requirements = ad.lookup("requirements")
        except KeyError:
            return False

        result = requirements.eval(scope=ad, target=self)
        if isinstance(result, Value):
            # Neither `undefined` nor `error` are `true`.
            return False;
        if isinstance(result, bool):
            return result
        if isinstance(result, int):
            return bool(result)
        if isinstance(result, float):
            return bool(result)

        return False;


    def printJson(self) -> str:
        """
        Serialize this ClassAd in the JSON format.
        """
        return _classad_print_json(self._handle)

    formatJson = printJson

    def printOld(self) -> str:
        """
        Serialize this ClassAd in the "old" format.
        """
        return _classad_print_old(self._handle)

    formatOld = printOld

    def symmetricMatch(self, ad : "ClassAd") -> bool:
        '''
        Equivalent to ``self.matches(ad) and ad.matches(self)``.

        :param ad:  The ClassAd to match for and against.
        '''
        return self.matches(ad) and ad.matches(self)


    # In version 1, we didn't allow Python to check if others knew how
    # to do equality comparisons against us, which was probably wrong.
    #
    # If don't define __eq__(), the one we get from the ABC Mapping
    # allows comparisons against other `Mapping`s, which we may or
    # may not actually want.
    def __eq__(self, other):
        if type(self) is type(other):
            return _exprtree_eq(self._handle, other._handle)
        else:
            return NotImplemented


    #
    # The MutableMapping abstract methods.
    #

    def __len__(self):
        return _classad_size(self._handle)


    def __getitem__(self, key):
        if not isinstance(key, str):
            raise TypeError("ClassAd keys are strings")
        return _classad_get_item(self._handle, key, True)


    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise TypeError("ClassAd keys are strings")
        if isinstance(value, dict):
            return _classad_set_item(self._handle, key, ClassAd(value))
        return _classad_set_item(self._handle, key, value)


    def __delitem__(self, key):
        if not isinstance(key, str):
            raise TypeError("ClassAd keys are strings")
        return _classad_del_item(self._handle, key)


    def __iter__(self):
        keys = _classad_keys(self._handle)
        for key in keys:
            yield key


def _convert_local_datetime_to_utc_ts(dt):
    if dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None:
        the_epoch = datetime(1970, 1, 1, tzinfo=dt.tzinfo)
        return (dt - the_epoch) // timedelta(seconds=1)
    else:
        the_epoch = datetime(1970, 1, 1)
        naive_ts = (dt - the_epoch) // timedelta(seconds=1)
        offset = dt.astimezone().utcoffset() // timedelta(seconds=1)
        return naive_ts - offset


def _parse_ads_generator(input, parser : Parser = Parser.Auto):
    file_offset = 0
    if not isinstance(input, str):
        file_offset = input.tell()

    input_string = "".join(input)
    string_offset = 0
    while True:
        (ad, offset) = _classad_parse_next(input_string, int(parser), string_offset)

        string_offset += offset
        if not isinstance(input, str):
            file_offset += offset
            input.seek(file_offset, 0)

        if ad is None or offset == 0:
            return

        yield ad


#
# Lines starting with '#' are ignored in the file parser.  The "old" and
# "new" parsers both handle C-style ('//') and C++-style ('/* */') comments,
# but not when a comment starts a new ad.
#
# The version 1 bindings didn't include JSON or XML, so I'm ignoring the
# fact that JSON doesn't allow multiple ads at all right now (and not
# testing XML).  We should fix this eventually, but that might be a long
# way away.
#
def _parseAds(input : Union[str, IO], parser : Parser = Parser.Auto) -> Iterator[ClassAd]:
    '''
    Returns a generator which will parse each ad in the input.

    Ads serialized in the :const:`ParserType.Old` format must be separated by blank lines.
    Ads serialized in the :const:`ParserType.New` format may be separated by blank lines.

    :param input:  One or more serialized ClassAds.  The serializations must
                   all be in the same format.
    :param parser:  Which parser to use (serialization format to assume).  If
                    unspecified, attempt to determine if the serialized ads
                    are in the :const:`ParserType.Old` or
                    :const:`ParserType.New` format.
    '''
    return _parse_ads_generator(input, parser)


def _parseOne(input : Union[str, Iterator[str]], parser : Parser = Parser.Auto) -> ClassAd:
    '''
    Parses all of the ads in the input, merges them into one, and returns it.

    If *input* is a single string,
    ads serialized in the :const:`ParserType.Old` format must be separated by blank lines;
    ads serialized in the :const:`ParserType.New` format may be separated by blank lines.

    If *input* is an iterator,
    each serialized ad must be in its own string.

    :param input:  One or more serialized ClassAds.  The serializations must
                   all be in the same format.
    :param parser:  Which parser to use (serialization format to assume).  If
                    unspecified, attempt to determine if the serialized ads
                    are in the :const:`ParserType.Old or
                    :const:`ParserType.New` formats.
    '''
    total_offset = 0
    if not isinstance(input, str):
        total_offset = input.tell()

    result = ClassAd()
    input_string = "".join(input)
    string_offset = 0
    while True:
        (firstAd, offset) = _classad_parse_next(input_string, int(parser), string_offset)

        if firstAd is None or offset == 0:
            return result

        if firstAd is not None:
            result.update(firstAd)
        string_offset += offset

        if not isinstance(input, str):
            total_offset += offset
            input.seek(total_offset, 0)


# In version 1, this never actually returned an iterator, and it only
# advanced `input` if it was not a string.  Not sure why we bothered
# to allow strings here at all, actually.
def _parseNext(input : Union[str, IO], parser : Parser = Parser.Auto) -> ClassAd:
    '''
    Parses the first ad in the input and returns it.

    Ads serialized in the :const:`ParserType.Old` format must be separated by blank lines.
    Ads serialized in the :const:`ParserType.New` format may be separated by blank lines.

    You must specify the parser type if *input* is not a string and
    can not be rewound.

    :param input:  One or more serialized ClassAds.  The serializations must
                   all be in the same format.
    :param parser:  Which parser to use (serialization format to assume).
    :raises ClassAdException:  If ``input`` is not a string and can
      not be rewound.
    '''
    if isinstance(input, str):
        (firstAd, offset) = _classad_parse_next(input, int(parser))
        return firstAd
    else:
        if parser is Parser.Auto:
            try:
                parser = Parser.New

                location = input.tell()
                while True:
                    c = input.read(1)
                    if c == "/" or c == "[":
                        break
                    if not c.isspace():
                        parser = Parser.Old
                        break
                    if len(c) == 0:
                        break
                input.seek(location)
            except io.UnsupportedOperation as e:
                raise ClassAdException("Can not look ahead in stream; you must explicitly specify a parser.") from e
        return _classad_parse_next_fd(input.fileno(), int(parser))


def _quote(input : str) -> str:
    '''
    Quote *input* so it can be used for building ClassAd expressions.

    :param input:  The string to quote according the ClassAd syntax.
    '''
    return _classad_quote(input)


def _unquote(input : str) -> str:
    '''
    The inverse of :func:`quote`.

    :param input:  The ClassAd string to unquote into its literal value.
    '''
    return _classad_unquote(input)


def _lastError() -> str:
    '''
    Return the string representation of the last error to occur in the
    ClassAd library.

    As the ClassAd language has no concept of an exception, this is the
    only mechanism to receive detailed error messages from functions.
    '''
    return _classad_last_error()


def _version() -> str:
    '''
    Returns the version of ClassAd library this module is linked against.
    '''
    return _classad_version()

# Sphinx whines about forward references in the type hints if this isn't here.
from ._expr_tree import ExprTree



#!/usr/bin/python3

import os
import platform
import shutil
import errno
import argparse
import sys


BASE_CONFIG_TEMPLATE = """
# Start with a standard personal condor configuration
# (This is based on make-personal-from-tarball)
RELEASE_DIR={prefix_path}
LOCAL_DIR = {pwd}/local_dir
LOCAL_CONFIG_DIR = $(LOCAL_DIR)/config.d
use security: user_based
use role: personal
CONDOR_HOST = $(FULL_HOSTNAME)

# Add some tweaks for the testing environment
NEGOTIATOR_CYCLE_DELAY=1
NEGOTIATOR_INTERVAL=1
NEGOTIATOR_MIN_INTERVAL=1
SCHEDD_INTERVAL=1
SCHEDD_MIN_INTERVAL=1
DAGMAN_USER_LOG_SCAN_INTERVAL = 1
MAIL={true_path}
SENDMAIL={true_path}
LOCAL_CONFIG_FILE={pwd}/condor_config.local
DOCKER_PERFORM_TEST=false
SINGULARITY_TEST_SANDBOX_TIMEOUT=8
# The default FILETRANSFER_PLUGINS contains support for box,onedrive,
# and gdrive.  On startup, the startd runs each of these with the -classad
# option to probe.  However, this is somewhat slow, as they are each implemented
# in python, and we test none of them.  Ideally, I'd like to just be able to
# remove those from the default, but we don't have a good way to do that.
FILETRANSFER_PLUGINS = $(LIBEXEC)/curl_plugin

# ctest allows tests to run for up to 1500 seconds; after that, we are
# potentially a runaway personal condor instance.
DAEMON_SHUTDOWN=time() - DaemonStartTime > 1500

# The param table assumes singularity/apptainer has been installed
# from the tarball into libexec.  Let's use the system onedrive

SINGULARITY = /usr/bin/singularity
{extra_args}
"""


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument("--test", help="Name of HTCondor test")
    parser.add_argument(
        "--working-dir",
        help="Name of top-level CMake binary directory",
        default=".",
        dest="working_dir",
    )
    parser.add_argument(
        "--source-dir",
        help="Name of top-level HTCondor source directory",
        default=".",
        dest="source_dir",
    )
    parser.add_argument(
        "--prefix-path",
        help="Location of the install prefix to test",
        default=".",
        dest="prefix_path",
    )
    parser.add_argument("--dependencies", help="Source file dependencies")
    parser.add_argument(
        "--java",
        help="Enable Java universe",
        dest="java",
        action="store_true",
        default=False,
    )

    return parser.parse_args()


def write_base_config(prefix_path, java=False):
    extra_args = ""
    if not java:
        extra_args += "JAVA="
    with open("condor_config.local", "w") as fp:
        pass
    os.umask(0o022)
    os.makedirs("local_dir")
    pwd = os.getcwd()
    with open("base_config", "w") as fp:
        fp.write(
            BASE_CONFIG_TEMPLATE.format(
                pwd=pwd, prefix_path=prefix_path, extra_args=extra_args,
                true_path="/usr/bin/true" if sys.platform == "darwin" else "/bin/true"
            )
        )


def main():
    args = parse_args()

    rundir_parent = os.path.abspath(
        os.path.join(args.working_dir, "src", "condor_tests")
    )
    
    rundir = os.path.abspath(
        os.path.join(rundir_parent, args.test + "_ctest")
    )

    # Ensure helper scripts are in place.
    for helper in ["run_test.pl", "Condor.pm", "CondorUtils.pm", "CondorPersonal.pm", "CondorTest.pm"]:
        if not os.path.exists(os.path.join(rundir_parent, helper)):
            shutil.copy(os.path.join(args.source_dir, "src", "condor_tests", helper), rundir_parent)

    if os.path.exists(rundir):
        shutil.rmtree(rundir)
    os.makedirs(rundir)
    os.chdir(rundir)

    base_src_fname = os.path.join(args.source_dir, "src", "condor_tests", args.test)
    possibilities = [
        base_src_fname + ".run",
        base_src_fname + ".py",
    ]
    for possibility in possibilities:
        if os.path.exists(possibility):
            src_fname = possibility
            break
    else:  # i.e., if we did not break, do this...
        raise Exception(
            "Test file named {} with suffix .run or .py not found!".format(args.test)
        )

    dst_fname = os.path.join(rundir, os.path.split(src_fname)[1])
    shutil.copy(src_fname, dst_fname)

    for src_fname in args.dependencies.split(";"):
        if not src_fname:
            continue
        fname = os.path.split(src_fname)[1]
        dst_fname = os.path.join(rundir, fname)
        src_fname = os.path.join(args.source_dir, src_fname)
        if os.path.isdir(src_fname):
            shutil.copytree(src_fname, dst_fname)
        else:
            shutil.copy(src_fname, dst_fname)

    # TODO: there is much environment munging and path construction below that won't work on Windows

    os.environ["PERL5LIB"] = os.pathsep.join(["..", "."])


    new_library_path = os.path.join(args.prefix_path, "lib")
    if "LD_LIBRARY_PATH" in os.environ:
        os.environ["LD_LIBRARY_PATH"] = (
            new_library_path + ":" + os.environ["LD_LIBRARY_PATH"]
        )
    else:
        os.environ["LD_LIBRARY_PATH"] = new_library_path

    # macOS clears DYLD_LIBRARY_PATH from the environment of system
    # programs (e.g. /usr/bin/perl), so we can't use that to have the
    # condor_tests binaries find the condor libraries.
    # Making a symlink in src/condor_tests is our best option here.
    if platform.system() == "Darwin":
        try:
            symlink_dst = os.path.join(args.working_dir, "src", "condor_tests", "lib")
            os.symlink(new_library_path, symlink_dst)
        except FileExistsError:
            pass

    os.environ["PATH"] = os.pathsep.join(
        [
            os.path.join(args.prefix_path, "bin"),
            os.path.join(args.prefix_path, "sbin"),
            
            "/bin",
            "/sbin",
            "/usr/sbin",
            "/usr/bin",
        ]
    )

    if platform.system() == "Windows":
        os.environ["PATH"] += ";" + os.path.join(args.prefix_path, "../msconfig")        
    
    
    pythonpath = os.environ.setdefault("PYTHONPATH", "")
    add_to_path = os.path.join(args.prefix_path, "lib", "python")
    if add_to_path not in pythonpath:
        os.environ["PYTHONPATH"] = os.pathsep.join([".", add_to_path, pythonpath])

    write_base_config(args.prefix_path, java=args.java)
    # This is not re-generated each time.
    if os.path.exists("derived_condor_config"):
        os.unlink("derived_condor_config")

    os.environ["CONDOR_CONFIG"] = os.path.abspath("base_config")

    command_line = "perl" + " " + os.path.join("..", "run_test.pl") + " " + args.test

    r = os.system(command_line)
    if platform.system() == "Windows":
        exit(r)
    else:
        exit(r >> 8)



if __name__ == "__main__":
    main()



#!/usr/bin/env pytest

import classad2
import htcondor2

from ornithology import (
    action,
    Condor,
    ClusterState,
);


@action
def the_condor(test_dir):
    local_dir = test_dir / "condor"

    with Condor(
        local_dir=local_dir,
        config={
            "STARTD_ENVIRONMENT":           ";http_proxy=;https_proxy=",
        },
    ) as the_condor:
        yield the_condor


@action
def expected_prl():
    # We can add more detail as warranted.
    return [
        {
            "TransferUrl": 'debug://error/resolution[FailedName="a.b.c"]',
        },
        {
            "TransferUrl": 'debug://error/resolution[FailedName="x.y.z"]',
        },
        {
            "TransferUrl": 'file://no-such.file/1',
        },
        {
            "TransferUrl": 'file://no-such.file/2',
        },
        {
            "TransferUrl": 'http://no-such.tld/3',
        },
        {
            "TransferUrl": 'http://no-such.tld/4',
        },
    ]


@action
def expected_pil():
    expected_debug = {
        "TransferClass":    1,
        "PluginBasename":   "debug_plugin.py",
        "Schemes":          "debug",
        "ExitBySignal":     False,
    }

    expected_curl = {
        "TransferClass":    1,
        "PluginBasename":   "curl_plugin",
        "Schemes":          "file,http",
        "ExitBySignal":     False,
    }

    return (expected_debug, expected_curl)


@action
def the_held_job(the_condor, path_to_sleep, test_dir):
    the_description = {
        "universe":                 "vanilla",
        "executable":               path_to_sleep,
        "arguments":                1,
        "transfer_executable":      False,
        "should_transfer_files":    True,
        "log":                      test_dir / "sleep.log",
        "request_cpus":             1,
        "request_memory":           1,
        "transfer_input_files":     ",".join([
            'file://no-such.file/1',
            'file://no-such.file/2',
            'debug://error/resolution[FailedName="a.b.c"]',
            'debug://error/resolution[FailedName="x.y.z"]',
            'http://no-such.tld/3',
            'http://no-such.tld/4',
        ]),
    }
    job_handle = the_condor.submit(the_description, count=1)

    assert job_handle.wait(
        condition=ClusterState.all_held,
        fail_condition=ClusterState.any_complete,
        timeout=60,
    )

    return job_handle


@action
def the_input_epoch(the_held_job, the_condor):
    # There should be a v2 binding for this.
    p = the_condor.run_command([
        "condor_history",
        "-epochs",
        "-type", "input",
        "-limit", 1,
        "-l",
        the_held_job.clusterid,
    ])
    ad = classad2.parseOne(p.stdout)

    return ad


@action
def actual_prl(the_input_epoch):
    return the_input_epoch['InputPluginResultList']


@action
def actual_pil(the_input_epoch):
    return the_input_epoch['InputPluginInvocations']


class TestMultipleSchemes:

    def test_plugin_result_list(self, actual_prl, expected_prl):
        assertions = 0
        attributes = 0

        for expected in expected_prl:
            attributes += len(expected)
            transferURL = expected["TransferUrl"]

            # We could break after the asserting that the values match, but
            # continuing also asserts that each ad in the actual PRL has
            # a TransferURL, which catches the case where we accidentally
            # stick a PIL in it.
            for actual in actual_prl:
                if actual["TransferUrl"] != transferURL:
                    continue
                for key in expected.keys():
                    assert actual[key] == expected[key]
                    assertions += 1

        assert assertions == attributes


    def test_plugin_invocation_list(self, actual_pil, expected_pil):
        assertions = 0
        attributes = 0

        # We could break after the asserting that the values match, but
        # continuing also asserts that each ad in the actual PIL has
        # a PluginBasename, which catches the case where we accidentally
        # stick a PRL in it.
        for expected in expected_pil:
            attributes += len(expected)
            plugin_basename = expected["PluginBasename"]
            for actual in actual_pil:
                if actual["PluginBasename"] != plugin_basename:
                    continue
                for key in expected.keys():
                    assert actual[key] == expected[key]
                    assertions += 1

        assert assertions == attributes



import os
import sys
import stat
import tempfile
import time
import shutil
import math
import enum

import htcondor2

from datetime import datetime, timedelta
from pathlib import Path
from htcondor2._utils.ansi import Color, colorize, bold, stylize, AnsiOptions

from typing import Union, List, Optional

class HistogramMode(enum.Enum):
    CUMULATIVE = enum.auto()
    INSTANT = enum.auto()

def produce_job_event_histogram(log_file: Union[Path, str], mode: HistogramMode) -> List[str]:
    log_file = Path(log_file)

    if not log_file.exists():
        raise FileNotFoundError(f"Could not find file: {str(log_file)}")

    if os.access(log_file, os.R_OK) is False:
        raise PermissionError(f"Could not access file: {str(log_file)}")

    # State -> (Label, Color)
    # NOTE: In stack order (done->held : bottom->top)
    # 256-color palette IDs for half-block rendering (fg/bg via AnsiOptions)
    KEY_LABEL = 0
    KEY_COLOR = 1
    STATES = {
        "done": ("Done", 2),            # green
        "running": ("Running", 6),      # cyan
        "xfer_in": ("Xfer In", 4),      # blue
        "xfer_out": ("Xfer Out", 5),    # magenta
        "idle": ("Idle", 3),            # yellow
        "held": ("Held", 1),            # red
    }

    def resolve_transfer_event(event) -> Optional[str]:
        """Determine state from transfer event"""
        XFER = {
            1: "xfer_in",      # IN_QUEUED
            2: "xfer_in",      # IN_STARTED
            3: "running",      # IN_FINISHED
            4: "xfer_out",     # OUT_QUEUED
            5: "xfer_out",     # OUT_STARTED
            6: "running",      # OUT_FINISHED
        }
        return XFER.get(event.get("Type"))

    # Map of job events to desired states
    EVENT_TO_STATE = {
        htcondor2.JobEventType.SUBMIT: "idle",
        htcondor2.JobEventType.EXECUTE: "running",
        htcondor2.JobEventType.JOB_HELD: "held",
        htcondor2.JobEventType.JOB_RELEASED: "idle",
        htcondor2.JobEventType.JOB_EVICTED: "idle",
        htcondor2.JobEventType.JOB_TERMINATED: "done",
        htcondor2.JobEventType.JOB_ABORTED: "done",
        htcondor2.JobEventType.JOB_STAGE_IN: "xfer_in",
        htcondor2.JobEventType.JOB_STAGE_OUT: "xfer_out",
        htcondor2.JobEventType.FILE_TRANSFER: resolve_transfer_event,
    }

    # List of lines to return (built from top -> bottom of terminal)
    LINES = list()

    # Parse events from the log file
    transitions = []
    with htcondor2.JobEventLog(str(log_file)) as event_log:
        for event in event_log.events(0):
            job_id = f"{event.cluster}.{event.proc}"
            timestamp = datetime.strptime(
                event.get("EventTime"), "%Y-%m-%dT%H:%M:%S"
            )

            res = EVENT_TO_STATE.get(event.type)
            new_state = res(event) if callable(res) else res

            if new_state is not None:
                transitions.append((timestamp, job_id, new_state))

    if not transitions:
        LINES.append("No state-change events found in log file.")
        return LINES

    transitions.sort(key=lambda t: t[0])

    t_start = transitions[0][0]
    t_end = transitions[-1][0]
    total_seconds = (t_end - t_start).total_seconds()

    if total_seconds == 0:
        LINES.append("All events have the same timestamp; cannot build histogram.")
        return LINES

    # Terminal dimensions
    term_size = shutil.get_terminal_size()
    term_width = term_size.columns
    term_height = term_size.lines

    y_label_width = 7   # "NNNNNN "
    chart_width = term_width - y_label_width - 1  # 1 for │
    chart_height = term_height - 7  # title, blank, x-axis, x-labels, blank, legend, new terminal line

    if chart_width < 10 or chart_height < 3:
        LINES.append("Terminal too small for histogram.")
        return LINES

    # Build time bins
    bin_seconds = total_seconds / chart_width
    bins = []
    event_idx = 0

    job_states = {}
    for col in range(chart_width):
        counts = {s: 0 for s in STATES}
        bin_end = t_start + timedelta(seconds=(col + 1) * bin_seconds)
        while event_idx < len(transitions) and transitions[event_idx][0] <= bin_end:
            _, jid, st = transitions[event_idx]
            if mode == HistogramMode.CUMULATIVE:
                job_states[jid] = st
            else:
                counts[st] += 1
            event_idx += 1
        if mode == HistogramMode.CUMULATIVE:
            for st in job_states.values():
                counts[st] += 1

        bins.append(counts)

    max_total = max(sum(b.values()) for b in bins)
    if max_total == 0:
        LINES.append("No jobs found in log file.")
        return LINES

    # Render title
    mode_label = "Cumulative" if mode == HistogramMode.CUMULATIVE else "Instant"
    LINES += [
        bold(f"DAG Event Log Histogram ({mode_label}): {log_file.name}"),
        ""
    ]

    # Compute nice Y-axis tick marks (~5 labels)
    def nice_num(x, do_round=True):
        """Round x to a 'nice' number (1, 2, 5 × 10^n)."""
        if x <= 0:
            return 1
        exp = math.floor(math.log10(x))
        frac = x / (10 ** exp)
        if do_round:
            if frac < 1.5:
                nice = 1
            elif frac < 3:
                nice = 2
            elif frac < 7:
                nice = 5
            else:
                nice = 10
        else:
            if frac <= 1:
                nice = 1
            elif frac <= 2:
                nice = 2
            elif frac <= 5:
                nice = 5
            else:
                nice = 10
        return nice * (10 ** exp)

    tick_step = nice_num(max_total / 5)
    if tick_step < 1:
        tick_step = 1
    tick_step = int(tick_step)

    # Build set of Y values that get labels
    row_labels = {}
    # Generate tick values
    tick_val = tick_step
    while tick_val <= max_total:
        # Map value to row: row = value * chart_height / max_total, rounded
        row = round(tick_val * chart_height / max_total)
        if 1 <= row <= chart_height:
            row_labels[row] = tick_val
        tick_val += tick_step
    # Always label top
    row_labels[chart_height] = max_total

    # Helper: find which state a given Y value falls in for a bin
    def state_at_y(b, y_val):
        """Return the state name at y_val in stacked bar b, or None if empty."""
        cum = 0
        for state in STATES:
            cum += b[state]
            if cum > y_val:
                return state
        return None

    # Render chart rows top to bottom using half-block characters
    for row in range(chart_height, 0, -1):
        # Two sub-rows per character cell
        top_threshold = (row - 0.25) * max_total / chart_height
        bot_threshold = (row - 0.75) * max_total / chart_height

        # Y-axis label (manually print 0 row flush with x-axis)
        if row in row_labels:
            label = f"{row_labels[row]:>6} "
        else:
            label = "       "

        line = label + "│"

        for col in range(chart_width):
            b = bins[col]
            top_state = state_at_y(b, top_threshold)
            bot_state = state_at_y(b, bot_threshold)

            if top_state is None and bot_state is None:
                # Both empty
                line += " "
            elif top_state == bot_state:
                # Both same state - full block
                line += stylize("█", AnsiOptions(color_id=STATES[top_state][KEY_COLOR]))
            elif top_state is None and bot_state is not None:
                # Only bottom half filled
                line += stylize("▄", AnsiOptions(color_id=STATES[bot_state][KEY_COLOR]))
            elif top_state is not None and bot_state is None:
                # Only top half filled
                line += stylize("▀", AnsiOptions(color_id=STATES[top_state][KEY_COLOR]))
            else:
                # Different states: ▄ with fg=bottom, bg=top
                line += stylize("▄", AnsiOptions(
                    color_id=STATES[bot_state][KEY_COLOR],
                    highlight=STATES[top_state][KEY_COLOR],
                ))

        LINES.append(line)

    # X-axis line
    LINES.append("     0 └" + "─" * chart_width)

    # Time labels
    if total_seconds < 86400:
        fmt = "%H:%M:%S"
    else:
        fmt = "%m/%d %H:%M"

    start_str = t_start.strftime(fmt)
    end_str = t_end.strftime(fmt)
    mid_time = t_start + timedelta(seconds=total_seconds / 2)
    mid_str = mid_time.strftime(fmt)

    label_chars = [" "] * chart_width

    # Place start label
    for i, c in enumerate(start_str):
        if i < chart_width:
            label_chars[i] = c

    # Place end label
    end_pos = chart_width - len(end_str)
    for i, c in enumerate(end_str):
        idx = end_pos + i
        if 0 <= idx < chart_width:
            label_chars[idx] = c

    # Place mid label only if it fits without overlapping
    mid_pos = chart_width // 2 - len(mid_str) // 2
    if mid_pos > len(start_str) + 1 and mid_pos + len(mid_str) < end_pos - 1:
        for i, c in enumerate(mid_str):
            idx = mid_pos + i
            if 0 <= idx < chart_width:
                label_chars[idx] = c

    LINES.append("        " + "".join(label_chars))

    # Legend with double-width swatches and wider spacing
    legend_parts = []
    for state in STATES:
        swatch = stylize("██", AnsiOptions(color_id=STATES[state][KEY_COLOR]))
        legend_parts.append(swatch + " " + STATES[state][KEY_LABEL])
    LINES += [
        "",
        "  " + "   ".join(legend_parts),
    ]

    return LINES


def print_job_event_histogram(log_file: Union[Path, str], mode: HistogramMode) -> None:
    for line in produce_job_event_histogram(log_file, mode):
        print(line)




#!/usr/bin/env pytest

import re
from time import sleep

from ornithology import *
from htcondor_cli.convert_ad import _ad_to_daemon_status

UNIT_TEST_CASES = {
    # Verify No MyType -> None
    "NO_MYTYPE" : {
        "foo" : "bar"
    },
    # Verify Wrong MyType -> None
    "FILTERED_MYTYPE" : {
        "MyType" : "Shadow"
    },
    # Verify MyType for master and 0 rdcdc -> 100 HP
    "MASTER_RDCDC_0%" : {
        "MyAddress" : "FILLER",
        "Machine" : "LOCAL_HOST",
        "MyType" : "DaemonMaster",
        "RecentDaemonCoreDutyCycle" : 0,
        "ExpectedDaemon" : "MASTER",
        "ExpectedHP" : 100,
    },
    # Verify MyType Collector and 1 rdcdc -> 0 HP
    "COLLECTOR_RDCDC_100%" : {
        "MyAddress" : "FILLER",
        "Machine" : "LOCAL_HOST",
        "MyType" : "Collector",
        "RecentDaemonCoreDutyCycle" : 1,
        "ExpectedDaemon" : "COLLECTOR",
        "ExpectedHP" : 0,
    },
    # Verify MyType Negotiator and 0.98 rdcdc -> 0 HP
    "NEGOTIATOR_RDCDC_98%" : {
        "MyAddress" : "FILLER",
        "Machine" : "LOCAL_HOST",
        "MyType" : "Negotiator",
        "RecentDaemonCoreDutyCycle" : 0.98,
        "ExpectedDaemon" : "NEGOTIATOR",
        "ExpectedHP" : 0,
    },
    # Verify MyType Startd and 0.95 rdcdc -> 95 HP
    "STARTD_RDCDC_95%" : {
        "MyAddress" : "FILLER",
        "Machine" : "LOCAL_HOST",
        "MyType" : "Startd",
        "RecentDaemonCoreDutyCycle" : 0.95,
        "ExpectedDaemon" : "STARTD",
        "ExpectedHP" : 95,
    },
    # Verify 0.33 rdcdc -> 98 HP
    "MASTER_RDCDC_33%" : {
        "MyAddress" : "FILLER",
        "Machine" : "LOCAL_HOST",
        "MyType" : "DaemonMaster",
        "RecentDaemonCoreDutyCycle" : 0.33,
        "ExpectedDaemon" : "MASTER",
        "ExpectedHP" : 100,
    },
    # Verify 0.96 rdcdc -> 63 HP
    "MASTER_RDCDC_96%" : {
        "MyAddress" : "FILLER",
        "Machine" : "LOCAL_HOST",
        "MyType" : "DaemonMaster",
        "RecentDaemonCoreDutyCycle" : 0.96,
        "ExpectedDaemon" : "MASTER",
        "ExpectedHP" : 63,
    },
    # Verify 0.97 rdcdc -> 31 HP
    "MASTER_RDCDC_97%" : {
        "MyAddress" : "FILLER",
        "Machine" : "LOCAL_HOST",
        "MyType" : "DaemonMaster",
        "RecentDaemonCoreDutyCycle" : 0.97,
        "ExpectedDaemon" : "MASTER",
        "ExpectedHP" : 31,
    },
    # Verify MyType Schedd and 0 rdcdc -> 100 HP
    "SCHEDD_RDCDC_0%" : {
        "MyAddress" : "FILLER",
        "Machine" : "LOCAL_HOST",
        "MyType" : "Scheduler",
        "RecentDaemonCoreDutyCycle" : 0,
        "ExpectedDaemon" : "SCHEDD",
        "ExpectedHP" : 100,
    },
    # Verify Schedd 1 rdcdc -> 28 HP
    "SCHEDD_RDCDC_100%" : {
        "MyAddress" : "FILLER",
        "Machine" : "LOCAL_HOST",
        "MyType" : "Scheduler",
        "RecentDaemonCoreDutyCycle" : 1,
        "ExpectedDaemon" : "SCHEDD",
        "ExpectedHP" : 28,
    },
    # Verify Schedd with 0 bytes per second Upload -> 93
    "SCHEDD_UPLOAD_0_BYTES_PER_SEC" : {
        "MyAddress" : "FILLER",
        "Machine" : "LOCAL_HOST",
        "MyType" : "Scheduler",
        "RecentDaemonCoreDutyCycle" : 0,
        "TransferQueueNumUploading" : 1,
        "FileTransferUploadBytesPerSecond_1m" : 0,
        "ExpectedDaemon" : "SCHEDD",
        "ExpectedHP" : 93,
    },
    # Verify Download MB Waiting < 100 -> 100 HP
    "SCHEDD_VALID_DOWLOAD_WAITING" : {
        "MyAddress" : "FILLER",
        "Machine" : "LOCAL_HOST",
        "MyType" : "Scheduler",
        "RecentDaemonCoreDutyCycle" : 0,
        "FileTransferMBWaitingToDownload" : 10,
        "TransferQueueDownloadWaitTime" : 100000,
        "TransferQueueNumWaitingToDownload" : 3,
        "ExpectedDaemon" : "SCHEDD",
        "ExpectedHP" : 100,
    },
    # Verify Download MB Waiting > 100 && Wait time < 5m -> 100 HP
    "SCHEDD_VALID_DOWLOAD_WAIT_TIME" : {
        "MyAddress" : "FILLER",
        "Machine" : "LOCAL_HOST",
        "MyType" : "Scheduler",
        "RecentDaemonCoreDutyCycle" : 0,
        "FileTransferMBWaitingToDownload" : 250,
        "TransferQueueDownloadWaitTime" : 20,
        "TransferQueueNumWaitingToDownload" : 3,
        "ExpectedDaemon" : "SCHEDD",
        "ExpectedHP" : 100,
    },
    # Verify Download MB Waiting > 100 && Wait time > 5m -> 93 HP
    "SCHEDD_INVALID_DOWNLOAD_WAIT_TIME" : {
        "MyAddress" : "FILLER",
        "Machine" : "LOCAL_HOST",
        "MyType" : "Scheduler",
        "RecentDaemonCoreDutyCycle" : 0,
        "FileTransferMBWaitingToDownload" : 250,
        "TransferQueueDownloadWaitTime" : 100000,
        "TransferQueueNumWaitingToDownload" : 3,
        "ExpectedDaemon" : "SCHEDD",
        "ExpectedHP" : 93,
    },
}

@action(params={name: name for name in UNIT_TEST_CASES})
def unit_tests(request):
    return (request.param, UNIT_TEST_CASES[request.param])

@action
def unit_test_name(unit_tests):
    return unit_tests[0]

@action
def unit_test_case(unit_tests):
    return unit_tests[1]

@action
def unit_test_result(unit_test_case):
    return _ad_to_daemon_status(unit_test_case)

#===============================================================================================
TEST_CASES = {
    "ACCESS_POINT" : (["htcondor", "ap", "status"], 5),
    "CENTRAL_MANAGER" : (["htcondor", "cm", "status"], 0),
    "SERVER" : (["htcondor", "server", "status"], 0),
}

@standup
def condor(test_dir):
    with Condor(local_dir=test_dir / "condor",
                config={
                    "COLLECTOR_UPDATE_INTERVAL" : "2",
                    "DAEMON_LIST" : "MASTER,COLLECTOR NEGOTIATOR,STARTD SCHEDD",
                }) as condor:
        yield condor

@action
def run_commands(condor):
    OUTPUT = {}
    for test, info in TEST_CASES.items():
        sleep(info[1])
        p = condor.run_command(info[0])
        OUTPUT[test] = p.stdout + p.stderr
    return OUTPUT

@action(params={name: name for name in TEST_CASES})
def test_commands(run_commands, request):
    return (request.param, run_commands[request.param])

#===============================================================================================
class TestHTCondorCLISytemStatus:
    def test_convert_ad_unit_tests(self, unit_test_name, unit_test_case, unit_test_result):
        if "MyType" not in unit_test_case or unit_test_case["MyType"] == "Shadow":
            assert unit_test_result is None
        else:
            assert unit_test_case["ExpectedDaemon"] == unit_test_result[0]
            assert unit_test_case["ExpectedHP"] == int(unit_test_result[1]["HealthPoints"])

    def test_htcondor_cli_sys_status(self, test_commands):
        # Ignore the first line (header)
        for line in test_commands[1].split("\n")[1:]:
            if line.strip() != "":
                # Assert all commands return good health for items
                assert re.match("^.+ Good", line, re.IGNORECASE)
                if test_commands[0] == "SERVER":
                    assert re.match("^(master|collector|negotiator|schedd|startd|overall) .+", line, re.IGNORECASE)



from typing import List, Optional

from ._common_imports import (
    classad,
    Collector,
    DaemonType,
    CredType,
    CredCheck,
)

from .htcondor2_impl import (
    _credd_do_store_cred,
    _credd_do_check_oauth_creds,
    _credd_run_credential_producer,
    _credd_get_oauth2_credential,
    HTCondorException,
)

from ._param import _param as param


class Credd():

    def __init__(self, location : Optional[classad.ClassAd] = None):
        '''
        The credd client.  Adds, deletes, and queries (legacy) passwords,
        user credentials (new passwords, Kerberos), and user service
        credentials (OAUth2).

        If you're trying to submit a job, see :meth:`Submit.issue_credentials`
        instead.

        :param location:  A ClassAd with a ``MyAddress`` attribute, such as
                          might be returned by :meth:`Collector.locate`.
        '''
        self._default = False

        if location is None:
            self._default = True

            c = Collector()
            # The version 1 documentation says that the we assume the
            # local schedd, but the implementation uses the local credd.
            location = c.locate(DaemonType.Credd)

        if not isinstance(location, classad.ClassAd):
            raise TypeError("location must be a ClassAd")

        self._addr = location['MyAddress']
        # We never actually use this for anything.
        # self._version = location['CondorVersion']


    def add_password(self, password : str, user : Optional[str] = None) -> None:
        '''
        Store the specified *password* in the credd for the specified *user*.

        :param password:  The password to store.
        :param user:  The user for whom to store the password.  If
                      :const:`None`, attempt to guess based on the
                      current process's effective user ID.
        '''
        if len(password) == 0:
            # This was HTCondorValueError in version 1.
            raise ValueError("password may not be empty")

        mode = self._STORE_CRED_LEGACY_PWD | self._GENERIC_ADD

        addr = self._addr
        if self._default:
            addr = None

        _credd_do_store_cred(addr, user, password, mode, None, None, -1)


    def delete_password(self, user : Optional[str] = None) -> bool:
        '''
        Delete the password stored in the credd for the specified *user*.

        :param user:  The user for whom to delete the password.  If
                      :const:`None`, attempt to guess based on the
                      current process's effective user ID.
        '''
        mode = self._STORE_CRED_LEGACY_PWD | self._GENERIC_DELETE

        addr = self._addr
        if self._default:
            addr = None

        result = _credd_do_store_cred(addr, user, None, mode, None, None, -1)
        return result == self._SUCCESS


    def query_password(self, user : Optional[str] = None) -> bool:
        '''
        Check if the credd has a password stored for the specified *user*.

        :param user:  The user for whom to retrieve the password.  If
                      :const:`None`, attempt to guess based on the
                      current process's effective user ID.
        '''
        # FIXME: Test this.  A return of FAILURE_NOT_FOUND did
        # not raise an exception in version 1.
        mode = self._STORE_CRED_LEGACY_PWD | self._GENERIC_QUERY

        addr = self._addr
        if self._default:
            addr = None

        result = _credd_do_store_cred(addr, user, None, mode, None, None, -1)
        return result == self._SUCCESS


    def add_user_cred(self, credtype : CredType, credential : bytes, user : Optional[str] = None ) -> None:
        '''
        Store the specified *credential* of the specified *credtype*
        in the credd for the specified *user*.

        :param credtype:  One of :const:`CredType.Password` and
                          :const:`CredType.Kerberos`.
        :param credential:  The credential to store.
        :param user:  The user for whom to store the credential.  If
                      :const:`None`, attempt to guess based on the
                      current process's effective user ID.
        '''
        mode = self._GENERIC_ADD
        if credtype == CredType.Password:
            mode |= credtype
        elif credtype == CredType.Kerberos:
            mode |= credtype | self._STORE_CRED_WAIT_FOR_CREDMON
        else:
            raise ValueError("invalid credtype")

        if credential is None:
            producer = param.get("SEC_CREDENTIAL_PRODUCER")
            if producer == "CREDENTIAL_ALREADY_STORED":
                # This raised HTCondorIOError in version 1.
                return
            if producer is None:
                raise HTCondorException("No credential supplied but SEC_CREDENTIAL_PRODUCER not set.")

            credential = _credd_run_credential_producer(producer)

        if credential is None:
            # This was HTCondorValueError in version 1
            raise ValueError("credential may not be empty")

        _credd_do_store_cred(self._addr, user, credential, mode, None, None, -1)


    def delete_user_cred(self, credtype : CredType, user : Optional[str] = None) -> None:
        '''
        Delete the credential of the specified *credtype*
        stored in the credd for the specified *user*.

        :param credtype:  One of :const:`CredType.Password` and
                          :const:`CredType.Kerberos`.
        :param user:  The user for whom to delete the credential.  If
                      :const:`None`, attempt to guess based on the
                      current process's effective user ID.
        '''
        mode = self._GENERIC_DELETE | credtype
        _credd_do_store_cred(self._addr, user, None, mode, None, None, -1)


    def query_user_cred(self, credtype : CredType, user : Optional[str] = None) -> str:
        '''
        Check if the credd has a credential of the specific *credtype*
        stored in the credd for the specified *user*.

        :param credtype:  One of :const:`CredType.Password` and
                          :const:`CredType.Kerberos`.
        :param user:  The user for whom to retrieve the credential.  If
                      :const:`None`, attempt to guess based on the
                      current process's effective user ID.
        :return:  The time the credential was last updated, or ``None``.
        '''
        mode = self._GENERIC_QUERY
        if credtype == CredType.Password:
            mode |= credtype
        elif credtype == CredType.Kerberos or credtype == CredType.OAuth:
            mode |= credtype | self._STORE_CRED_WAIT_FOR_CREDMON
        else:
            # This was HTCondorEnumError in version 1.
            raise ValueError("invalid credtype")

        result = _credd_do_store_cred(self._addr, user, None, mode, None, None, -1)

        # This function returns "The time the credential was last updated,
        # or None."  6 is the SUCCESS_PENDING error code, which is not
        # considered a failure and therefore doesn't trigger a None return
        # from the C code.
        if result == 6:
            return None
        else:
            return result


    def add_user_service_cred(self, credtype : CredType, credential : bytes, service : str, handle : Optional[str] = None, user : Optional[str] = None, refresh : Optional[bool] = None) -> None:
        '''
        Store the specified OAuth *credential* in the credd for the specified *user*.

        :param credtype:  Must be :const:`CredType.OAuth`.
        :param credential:  The credential to store.
        :param service:  An identifier.  Used in submit files to map from
                         URLs to credentials.
        :param handle:  An optional identifier.  Used in submit files to
                        distinguish multiple credentials from the same
                        service.
        :param user:  The user for whom to store the credential.  If
                      :const:`None`, attempt to guess based on the
                      current process's effective user ID.
        :param refresh:  Is credential a refresh-style token that must be
                         processed by a CredMon to produce an access token.
                         If :const:`None`, have credd decide.
        '''
        mode = self._GENERIC_ADD | CredType.OAuth
        if credtype != CredType.OAuth:
            # This raised HTCondorIOError in version 1.
            raise ValueError("invalid credtype")

        if credential is None:
            producer = param.get("SEC_CREDENTIAL_PRODUCER")
            if producer == "CREDENTIAL_ALREADY_STORED":
                # This was HTCondorIOError in version 1.
                return
            if producer is None:
                raise HTCondorException("No credential supplied but SEC_CREDENTIAL_PRODUCER not set.")

            credential = _credd_run_credential_producer(producer)

        if credential is None:
            # This was HTCondorValueError in version 1
            raise ValueError("credential may not be empty")

        if refresh is None:
            refresh = -1

        _credd_do_store_cred(self._addr, user, credential, mode, service, handle, refresh)


    def delete_user_service_cred(self, credtype : CredType, service : str, handle : Optional[str] = None, user : Optional[str] = None) -> None:
        '''
        Delete the OAuth credential stored in the credd for the specified *user*.

        :param credtype:  Must be :const:`CredType.OAuth`.
        :param service:  An identifier.  Used in submit files to map from
                         URLs to credentials.
        :param handle:  An optional identifier.  Used in submit files to
                        distinguish multiple credentials from the same
                        service.
        :param user:  The user for whom to store the credential.  If
                      :const:`None`, attempt to guess based on the
                      current process's effective user ID.
        '''
        mode = self._GENERIC_DELETE | CredType.OAuth
        if credtype != CredType.OAuth:
            # This was HTCondorEnumError in version 1.
            raise ValueError("invalid credtype")

        _credd_do_store_cred(self._addr, user, None, mode, service, handle, -1)


    # In version 1, this returned a CredStatus instance, but the CredStatus
    # class defined only __str__(), and no function in the API accepted a
    # CredStatus as an argument.
    def query_user_service_cred(self, credtype : CredType, service : str, handle : Optional[str] = None, user : Optional[str] = None) -> str:
        '''
        Check if the credd has an OAuth credential stored for the
        specified *user*.

        :param credtype:  Must be :const:`CredType.OAuth`.
        :param service:  An identifier.  Used in submit files to map from
                         URLs to credentials.
        :param handle:  An optional identifier.  Used in submit files to
                        distinguish multiple credentials from the same
                        service.
        :param user:  The user for whom to store the credential.  If
                      :const:`None`, attempt to guess based on the
                      current process's effective user ID.
        :return:  The result ClassAd serialized in the "old" syntax.
        '''
        mode = self._GENERIC_QUERY | CredType.OAuth
        if credtype != CredType.OAuth:
            # This was HTCondorEnumError in version 1.
            raise ValueError("invalid credtype")

        return _credd_do_store_cred(self._addr, user, None, mode, service, handle, -1)


    # We should probably change the type of `services`, since they're not
    # defined by ClassAds anywhere else in the Python API.
    def check_user_service_creds(self, credtype : CredType, serviceAds : List[classad.ClassAd], user : Optional[str] = None) -> CredCheck:
        '''
        :param credtype:  Must be :const:`CredType.OAuth`.
        :param serviceAds:  A list of ClassAds specifying which services
                            are needed.
        :param user:  The user for whom to store the credential.  If
                      :const:`None`, attempt to guess based on the
                      current process's effective user ID.
        :raises HTCondorException:  If a problem occurs talking to the credd.
        '''

        mode = self._GENERIC_CONFIG | CredType.OAuth
        if credtype != CredType.OAuth:
            # This was HTCondorEnumError in version 1.
            raise ValueError("invalid credtype")

        services = []
        for ad in serviceAds:
            if not isinstance(ad, ClassAd):
                raise TypeError("service must be of type classad.ClassAd")

            name = ad.get("Service", None)
            # This was HTCondorValueError in version 1.
            raise ValueError("request has no 'Service' attribute")

            handle = ad.get("Handle", None)
            if handle is not None:
                name = f"{name}*{handle}"

            services.append(name)

        (result, url) = _credd_do_check_oauth_creds(self._addr, user, mode, serviceAds)

        # These were HTCondorIOError in version 1.
        if result == -1:
            raise HTCondorException("invalid services argument")
        elif result == -2:
            raise HTCondorException("could not locate CredD")
        elif result == -3:
            raise HTCondorException("startCommand failed")
        elif result == -4:
            raise HTCondorException("communication failure")
        elif result <= -5:
            raise HTCondorException("internal error")
        else:
            return CredCheck(",".join(services), url)


    def get_oauth2_credential(self, service : str, handle : str, user : Optional[str] = None) -> str:
        '''
        Return the specified OAuth2 credential.

        :param service:  An identifier.  Used in submit files to map from
                         URLs to credentials.
        :param handle:  An optional identifier.  Used in submit files to
                        distinguish multiple credentials from the same
                        service.
        :param user:  The user for whom to store the credential.  If
                      :const:`None`, attempt to guess based on the
                      current process's effective user ID.
        :return:  The requested credential.
        '''
        return _credd_get_oauth2_credential(self._addr, user, service, handle)


    _STORE_CRED_USER_KRB         = 0x20
    _STORE_CRED_USER_PWD         = 0x24
    _STORE_CRED_USER_OAUTH       = 0x28
    _STORE_CRED_LEGACY           = 0x40
    _STORE_CRED_WAIT_FOR_CREDMON = 0x80
    # Pretty sure the trailing | 0x40 is superfluous.
    _STORE_CRED_LEGACY_PWD       = _STORE_CRED_LEGACY | _STORE_CRED_USER_PWD | 0x40

    _GENERIC_ADD        = 0
    _GENERIC_DELETE     = 1
    _GENERIC_QUERY      = 2
    _GENERIC_CONFIG     = 3

    _SUCCESS            = 1
    _FAILURE            = 0



import os
from htc_helpers import *

DAG_CMDS = []

def find_dag_cmds(dir: str):
    dag_cmds = []
    dag_ref = os.path.join(dir, "automated-workflows", "dagman-reference.rst")
    with open(dag_ref, "r", encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            for cmd in get_all_defined_role("dag-cmd-def", line):
                if cmd not in dag_cmds:
                    dag_cmds.append(cmd)
    dag_cmds.sort()
    return dag_cmds

def dump(obj):
    for attr in dir(obj):
        print("obj.%s = %r" % (attr, getattr(obj, attr)))

def dagcom_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    root_dir = root_dir = get_rel_path_to_root_dir(inliner)[:-1]
    cmd_name, cmd_index = custom_ext_parser(text)
    if cmd_name not in DAG_CMDS:
        docname = inliner.document.settings.env.docname
        warn(f"{docname}:{lineno} | '{cmd_name}' DAG command not in defined list. Either a typo or not defined.")
    ref_link = f"href=\"{root_dir}/automated-workflows/dagman-reference.html#" + str(cmd_name) + "\""
    return make_ref_and_index_nodes(name, cmd_name, cmd_index,
                                    ref_link, rawtext, inliner, lineno, options)

def setup(app):
    global DAG_CMDS
    DAG_CMDS = find_dag_cmds(app.srcdir)
    app.add_role("dag-cmd", dagcom_role)




# -*- coding: utf-8 -*-
#
# Read the Docs Template documentation build configuration file, created by
# sphinx-quickstart on Tue Aug 26 14:19:49 2014.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# If this is not inserted at the beginning of the list, our 'jira' ticket
# extension is hidden by the system 'jira' API, if it exists.
sys.path.insert(0, os.path.abspath('extensions'))
sys.path.append(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'src', 'condor_tests'))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
# Note: This setting is only for man page build see same variable in
#       full_conf.py for html excludes
extensions = [
    'ticket',
    'macro',
    'macro-def',
    'subcom',
    'subcom-def',
    'index',
    'jira',
    'tool',
    'ad-attr',
    'hidden',
    'dag-cmd',
    'man-page-dummies',
    'fake-definition',
]

# Add any paths that contain templates here, relative to this directory.
#templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
# Note: This setting is only for man page build see same variable in
#       full_conf.py for html excludes
master_doc = 'man-pages/index'

# General information about the project.
project = u'HTCondor Manual'
copyright = u'1990-2026, Center for High Throughput Computing, Computer \
Sciences Department, University of Wisconsin-Madison, Madison, WI, US. \
Licensed under the Apache License, Version 2.0.'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
release = '25.8.2'

rst_epilog = """
.. |release_date| replace:: April 16, 2026

.. role:: config(code)
    :language: condor-config
.. role:: classad(code)
    :language: condor-classad
.. role:: ad-expr(code)
    :language: condor-classad-expr
.. role:: jdl(code)
    :language: condor-submit
.. role:: dag(code)
    :language: condor-dagman
"""

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# Note: This setting is only for man page build see same variable in
#       full_conf.py for html excludes
exclude_patterns = [f for f in os.listdir() if f != "man-pages"]

# The reST default role (used for this markup: `text`) to use for all
# documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'colorful'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# -- Options for manual page output ---------------------------------------

# Short list of man pages to build which is used to add generice info
MAN_PAGES = [
        ('classad_eval', 1),
        ('classads', 1),
        ('condor_adstash', 1),
        ('condor_advertise', 1),
        ('condor_check_password', 1),
        ('condor_check_userlogs', 1),
        ('condor_chirp', 1),
        ('condor_configure', 1),
        ('condor_config_val', 1),
        ('condor_continue', 1),
        ('condor_dagman', 1),
        ('condor_dag_checker', 1),
        ('condor_drain', 1),
        ('condor_evicted_files', 1),
        ('condor_fetchlog', 1),
        ('condor_findhost', 1),
        ('condor_gather_info', 1),
        ('condor_gpu_discovery', 1),
        ('condor_history', 1),
        ('condor_hold', 1),
        ('condor_install', 1),
        ('condor_job_router_info', 1),
        ('condor_master', 1),
        ('condor_now', 1),
        ('condor_off', 1),
        ('condor_on', 1),
        ('condor_ping', 1),
        ('condor_pool_job_report', 1),
        ('condor_power', 1),
        ('condor_preen', 1),
        ('condor_prio', 1),
        ('condor_procd', 1),
        ('condor_q', 1),
        ('condor_qusers', 1),
        ('condor_qedit', 1),
        ('condor_qsub', 1),
        ('condor_reconfig', 1),
        ('condor_release', 1),
        ('condor_remote_cluster', 1),
        ('condor_reschedule', 1),
        ('condor_restart', 1),
        ('condor_rm', 1),
        ('condor_rmdir', 1),
        ('condor_router_history', 1),
        ('condor_router_q', 1),
        ('condor_router_rm', 1),
        ('condor_run', 1),
        ('condor_set_shutdown', 1),
        ('condor_sos', 1),
        ('condor_ssh_start', 1),
        ('condor_ssh_to_job', 1),
        ('condor_ssl_fingerprint', 1),
        #('condor_stats', 1),
        ('condor_status', 1),
        ('condor_store_cred', 1),
        ('condor_submit', 1),
        ('condor_submit_dag', 1),
        ('condor_suspend', 1),
        ('condor_tail', 1),
        ('condor_test_token', 1),
        ('condor_token_create', 1),
        ('condor_token_fetch', 1),
        ('condor_token_list', 1),
        ('condor_token_request', 1),
        ('condor_token_request_approve', 1),
        ('condor_token_request_auto_approve', 1),
        ('condor_token_request_list', 1),
        ('condor_top', 1),
        ('condor_transfer_data', 1),
        ('condor_transform_ads', 1),
        ('condor_update_machine_ad', 1),
        ('condor_updates_stats', 1),
        ('condor_upgrade_check', 1),
        ('condor_urlfetch', 1),
        ('condor_userlog', 1),
        ('condor_userprio', 1),
        ('condor_vacate', 1),
        ('condor_vacate_job', 1),
        ('condor_version', 1),
        ('condor_wait', 1),
        ('condor_watch_q', 1),
        ('condor_who', 1),
        ('get_htcondor', 1),
        ('gidd_alloc', 1),
        ('htcondor', 1),
        ('procd_ctl', 1)
]

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
# e.g. ('man-pages/condor_chirp', 'condor_chirp', u'HTCondor Manual', [u'HTCondor Team'], 1)
man_pages = [
    (f'man-pages/{item[0]}', item[0], u'HTCondor Manual', [u'HTCondor Team'], item[1]) for item in MAN_PAGES
]

# If true, show URL addresses after external links.
# man_show_urls = False

# -- conf.py contains common configuration and man pages configuration
# -- full_conf.py contains configuration for the whole manual
sys.path.append(os.path.dirname(__file__))
MANPAGES = os.environ.get('MANPAGES') == 'True'
if not MANPAGES:
    from full_conf import *



from typing import (
    List,
    Optional,
)

from ._common_imports import (
    classad,
    Collector,
    DaemonType,
)

from .htcondor2_impl import (
    _placement_user_login,
    _placement_query_users,
    _placement_query_tokens,
    _placement_query_authorizations,
)


class Placementd():
    """
    The placementd client. Talk with the placementd to request and query about
    placement tokens for an AP.
    """

    def __init__(self, location : Optional[classad.ClassAd] = None):
        """
        :param location:  A ClassAd with a ``MyAddress`` attribute, such as
            might be returned by :meth:`htcondor2.Collector.locate`.  :py:obj:`None` means the
            the local daemon.
        """
        if location is None:
            c = Collector()
            location = c.locate(DaemonType.Placementd)

        if not isinstance(location, classad.ClassAd):
            raise TypeError("location must be a ClassAd")

        self._addr = location['MyAddress']
        # We never actually use this for anything.
        # self._version = location['CondorVersion']


    def userLogin(self,
        username : str,
        authorizations : str,
        project : Optional[str] = None,
        requester : Optional[str] = None,
    ) -> classad.ClassAd:
        """
        Request a user placement token.

        :param username:  Foreign identifier of the user who'll use the token.
        :param authorizations:  What authorizations token should grant.
        :param project:  What AP porject should jobs be tied to
        :param requester:  Foreign identifier of user making the request.
            :py:obj:`None` means the same value as username.

        :return:  A ClassAd with the token and additional information.
        """
        return _placement_user_login(self._addr, username, authorizations, project, requester)

    def queryUsers(self,
        username : Optional[str] = None,
    ) -> List[classad.ClassAd]:
        """
        Query information about users who are currently authorized.

        :param username:  Foreign dentifier of a specific user.
            :py:obj:`None` will return information about all users.

        :return:  A list ClassAds with information about each user.
        """
        return _placement_query_users(self._addr, username)

    def queryTokens(self,
        username : Optional[str] = None,
        valid_only : bool = False,
        token_id : Optional[str] = None,
    ) -> List[classad.ClassAd]:
        """
        Query information about issued tokens.

        :param username:  Foreign dentifier of a specific user.
            :py:obj:`None` will return information about tokens for all users.
        :param valid_only:  If :py:obj:`True`, return information only
            about tokens that haven't expired.
        :param token_id:  Token unique id (i.e. jti claim).
            If not :py:obj:`None`, return information only about the token
            with the given id.

        :return:  A list ClassAds with information about each token.
        """
        return _placement_query_tokens(self._addr, username, valid_only, token_id)

    def queryAuthorizations(self,
        username : Optional[str] = None,
    ) -> List[classad.ClassAd]:
        """
        Query information about defined authorizations.

        :param username:  Foreign dentifier of a specific user.
            If not :py:obj:`None`, return information only about
            authorizations granted to the specified user.

        :return:  A list ClassAds with information about each user.
        """
        return _placement_query_authorizations(self._addr, username)




#!/usr/bin/env pytest

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

from ornithology import (
    action,
    Condor,
    ClusterState,
    DaemonLog,
)


@action
def the_condor(test_dir):
    local_dir = test_dir / "condor"

    with Condor(
        local_dir=local_dir,
        config={
            "STARTD_DEBUG":     "D_SUB_SECOND D_CATEGORY D_FULLDEBUG",
            "SHADOW_DEBUG":     "D_SUB_SECOND D_CATEGORY D_FULLDEBUG",
            "STARTER_DEBUG":    "D_SUB_SECOND D_CATEGORY D_FULLDEBUG",
        },
    ) as condor:
        yield condor


@action
def the_job_description(test_dir):
    print("the_job_description(): enter")
    (test_dir / "input.txt").write_text("aBcDeF 98765")

    return {
        "LeaveJobInQueue":              True,
        "should_transfer_files":        True,
        "when_to_transfer_output":      "ON_EXIT",
        "transfer_input_files":         "input.txt",
        "transfer_output_files":        "output.txt",

        "output":                       test_dir / "the_job.out",
        "error":                        test_dir / "the_job.err",
        "log":                          test_dir / "the_job.log",

        "shell":                        "cat input.txt"
    }


@action
def the_job_handle(the_condor, the_job_description):
    print("the_job_handle(): enter")
    job_handle = the_condor.submit(
        description=the_job_description,
        count=1,
    )

    assert job_handle.wait(
        timeout=60,
        condition = ClusterState.all_held,
    )

    yield job_handle

    job_handle.remove()


class TestOutputFailure:

    def test_shadow_log(self, the_job_handle, the_condor):
        shadow_log = the_condor.shadow_log.open()
        lines = list(filter(
            lambda line: 'condor_read(): timeout reading' in line,
            shadow_log.read(),
        ))
        assert len(lines) == 0


    def test_starter_log(self, the_job_handle, the_condor, test_dir):
        # See test_guidance_commands for how to make this reliable
        # in the presence of multiple jobs.
        starter_log_path= test_dir / "condor" / "log" / f"StarterLog.slot1_1"
        starter_log = DaemonLog(starter_log_path).open()

        lines = list(filter(
            lambda line: 'Connection to shadow may be lost' in line,
            starter_log.read(),
        ))
        assert len(lines) == 0



#!/usr/bin/env pytest

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

import time
from ornithology import *

import htcondor2


@action
def the_condor(test_dir):
    local_dir = test_dir / "condor"

    with Condor(
        local_dir=local_dir,
        config={
            "DAEMON_LIST":
                "TEST_AWAITABLE_DEADLINE_SOCKETD COLLECTOR",
            "TEST_AWAITABLE_DEADLINE_SOCKETD":
                "$(LIBEXEC)/test_awaitable_deadline_socketd",
            # Ornithology needs this, for some reason.
            "TEST_AWAITABLE_DEADLINE_SOCKETD_LOG":
                f"$(LOG)/AwaitableDeadlineSocketLog",
            "DC_DAEMON_LIST":
                "+ TEST_AWAITABLE_DEADLINE_SOCKETD",
        },
    ) as condor:
        yield condor


@action
def the_hot_log(the_condor):
    with the_condor.use_config():
        libexec = htcondor2.param['LIBEXEC']
    the_log = the_condor._get_daemon_log('TEST_AWAITABLE_DEADLINE_SOCKETD')
    for line in the_log.open().read():
        if line.message.startswith('DaemonCore: command socket at '):
            [prefix, sinful] = line.message.split('DaemonCore: command socket at ')
            result = the_condor.run_command(
                [f'{libexec}/test_awaitable_deadline_socket_client', '-hot', sinful]
            )
            assert result.returncode == 0
            return the_log
    assert False


@action
def the_timeout_log(the_condor):
    with the_condor.use_config():
        libexec = htcondor2.param['LIBEXEC']
    the_log = the_condor._get_daemon_log('TEST_AWAITABLE_DEADLINE_SOCKETD')
    for line in the_log.open().read():
        if line.message.startswith('DaemonCore: command socket at '):
            [prefix, sinful] = line.message.split('DaemonCore: command socket at ')
            result = the_condor.run_command(
                [f'{libexec}/test_awaitable_deadline_socket_client', '-timeout', sinful]
            )
            assert result.returncode == 0
            return the_log
    assert False


class TestDCStdFunction:

    def test_hot_socket(self, the_condor, the_hot_log):
        # I'm not at all sure I can just re-open the_hot_log.
        the_log = the_condor._get_daemon_log('TEST_AWAITABLE_DEADLINE_SOCKETD')
        for line in the_log.open().read():
            print(line)
            if line.message == "[hot] Exchange complete.":
                return
        assert False


    def test_timeout(self, the_condor, the_timeout_log):
        # I'm not at all sure I can just re-open the_timeout_log.
        the_log = the_condor._get_daemon_log('TEST_AWAITABLE_DEADLINE_SOCKETD')
        for line in the_log.open().read():
            print(line)
            if line.message == "[timeout] Timed out waiting for third string.":
                return
        assert False



#!/usr/bin/env pytest

from pathlib import Path
from getpass import getuser

from ornithology import (
    config,
    standup,
    action,
    Condor,
)

import htcondor2
htcondor2.enable_debug()

import logging


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


TEST_CASES = {
    "local_issuer": {
        "config": {
            "SEC_CREDENTIAL_DIRECTORY_OAUTH":       "$(LOCAL_DIR)/oauth_credentials",
            "CREDMON_OAUTH_LOG":                    "$(LOG)/CredMonOAuthLog",
            "DAEMON_LIST":                          "$(DAEMON_LIST),CREDMON_OAUTH",
            "AUTO_INCLUDE_CREDD_IN_DAEMON_LIST":    "True",
            "TRUST_CREDENTIAL_DIRECTORY":           "True",
            "LOCAL_CREDMON_PROVIDER_NAME":          "scitokens",
            "LOCAL_CREDMON_TOKEN_AUDIENCE":         "https://localhost",
            "ALLOW_DAEMON":                         "*",
            "LOCAL_CREDMON_PUBLIC_KEY":             "$(LOCAL_DIR)/trust_domain_ca_public.pem",
            "LOCAL_CREDMON_PRIVATE_KEY":            "$(LOCAL_DIR)/trust_domain_ca_privkey.pem",
            "CREDD_DEBUG":                          "D_FULLDEBUG",
        },
        "the_directory":    "{local_dir}/oauth_credentials",
        "the_local_filename":     "scitokens.top",
        "the_plain_filename":     "plain.use",
    },
}


@action(params={name: name for name in TEST_CASES})
def the_test_tuple(request):
    return (request.param, TEST_CASES[request.param])


@action
def the_test_name(the_test_tuple):
    return the_test_tuple[0]


@action
def the_test_case(the_test_tuple):
    return the_test_tuple[1]


@action
def the_local_dir(the_test_name, test_dir):
    return Path(str(test_dir)) / the_test_name


@action
def the_directory(the_test_case, the_local_dir):
    return the_test_case['the_directory'].format(local_dir=the_local_dir)


@action
def the_username():
    # Seems like we should have a way to get this from HTCondor.
    return getuser()


@action
def the_local_filename(the_test_case):
    return the_test_case['the_local_filename']


@action
def the_plain_filename(the_test_case):
    return the_test_case['the_plain_filename']


@action
def fake_user_token_file(test_dir):
    token_file = test_dir / "the_fake_use_token"
    token_file.write_text("fake-user-token")
    return token_file


@action
def the_condor(the_test_case, the_local_dir):
    with Condor(
        local_dir=the_local_dir,
        config=the_test_case['config']
    ) as the_condor:
       yield the_condor


class TestIssueCredentials:

    def test_top_file_created(self, the_test_name, the_condor, the_directory, the_username, the_local_filename):
        # Strictly speaking, this would be a set-up error.
        assert not (Path(the_directory) / the_username / the_local_filename).exists()

        # Strictly speaking, this should be an @action.
        submit = htcondor2.Submit(
                f"""
                executable = /bin/sleep
                transfer_executable = false
                arguments = 5

                log = {the_test_name}.log
                """
        )
        with the_condor.use_config():
            submit.issue_credentials()

        # Check for a .use file as well?
        assert (Path(the_directory) / the_username / the_local_filename).exists()

    # Test that a user-added credential that's not managed by any credmons
    # results in a .use file in the credd's storage.
    def test_use_file_created(self, the_test_name, the_condor, the_directory, the_username, the_plain_filename, fake_user_token_file):
        # Strictly speaking, this would be a set-up error.
        assert not (Path(the_directory) / the_username / the_plain_filename).exists()

        rv = the_condor.run_command(
            ['htcondor', '-v', 'credential', 'add', 'oauth2',
             str(fake_user_token_file),
             '--service', 'plain'],
        )
        assert rv.returncode == 0
        logger.info(rv.stdout)
        logger.info(rv.stderr)

        assert (Path(the_directory) / the_username / the_plain_filename).exists()



#!/usr/bin/env pytest

#   test_dagman_inline_desc.py
#
#   This test that inline submit description
#   have the following behavior:
#       A. Ordering doesn't matter (i.e. can occur post JOB command)
#       B. Inlcude files can 

from ornithology import *
import htcondor2
import json

DAG_FILE = "test.dag"

#------------------------------------------------------------------
@action
def write_include_file(test_dir, path_to_sleep):
    include = test_dir / "dag.include"
    with open(str(include), "w") as f:
        f.write(
f"""
SUBMIT-DESCRIPTION INC @=desc
    executable = {path_to_sleep}
    arguments  = 0
@desc
""")
    return include

#------------------------------------------------------------------
@action
def write_splice(test_dir, path_to_sleep):
    splice = test_dir / "splice.dag"
    with open(str(splice), "w") as f:
        f.write(
f"""
SUBMIT-DESCRIPTION foo @=desc
    executable = {path_to_sleep}
    arguments  = 0
@desc

# Note: This should not take effect (or test will fail)
SUBMIT-DESCRIPTION sleep @=desc
    executable = {path_to_sleep}
    arguments  = 1000000
@desc

JOB A foo
JOB B sleep
JOB C @=desc
    executable = {path_to_sleep}
    arguments  = 0
@desc
""")
    return splice

#------------------------------------------------------------------
@action
def write_dag(test_dir, path_to_sleep, write_include_file, write_splice):
    dag = test_dir / DAG_FILE
    with open(str(dag), "w") as f:
        f.write(
f"""
JOB BEFORE sleep

SUBMIT-DESCRIPTION sleep @=desc
    executable = {path_to_sleep}
    arguments  = 0
@desc

JOB AFTER sleep
JOB INCLUDE INC

SPLICE A {write_splice}

INCLUDE {write_include_file}
""")
    return dag

#------------------------------------------------------------------
@action
def run_dag(request, default_condor, write_dag):
    dag = htcondor2.Submit.from_dag(str(write_dag))
    return default_condor.submit(dag)

#==================================================================
class TestDAGManInlinePosition:
    def test_dag_ran(self, run_dag):
        assert run_dag.wait(condition=ClusterState.all_complete, timeout=120)
        with open(DAG_FILE + ".metrics", "r") as f:
            stats = json.load(f)
            assert stats["exitcode"] == 0



#!/usr/bin/env pytest

from ornithology import (
    action,
    ClusterState,
)

import htcondor2 as htcondor

import os

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@action
def the_condor(default_condor):
    return default_condor


TEST_CASES = {
    "latemat": {
        "args": "20",

        "job": """
            args=30
            My.Pos = \"{$(Row),$(Step)}\"
            hold = 1
            max_materialize = 3

            queue my.foo,my.bar from (
                1 2
                3 4
                a z
            )
        """,

        "digest": """FACTORY.Requirements=MY.Requirements
FACTORY.Iwd={IWD}
hold=1
max_materialize=3
My.Pos=\"{{$(Row),$(Step)}}\"

Queue 1 my.foo,my.bar from """,

        "values": {
            0: {
                'Args': "20",
                'Pos': '{0,0}',
            },
            1: {
                'Args': "20",
                'Pos': '{1,0}',
            },
            2: {
                'Args': "20",
                'Pos': '{2,0}',
            },
        },
    },

    "latemat2": {
        "args": '30 -f $(foo) -b $(bar) -z $(baz)',

        "job": """
            args=30
            My.Pos = "{$(Row),$(Step),$(foo)}"
            hold = 1
            max_materialize = 3

            queue foo,bar from (
                1 2
                3 4
                a z
            )
        """,

        "digest": """FACTORY.Requirements=MY.Requirements
args=30 -f $(foo) -b $(bar) -z 
FACTORY.Iwd={IWD}
hold=1
max_materialize=3
My.Pos=\"{{$(Row),$(Step),$(foo)}}\"

Queue 1 foo,bar from """
        ,

        "values": {
            0: {
                'Args': "30 -f 1 -b 2 -z",
                'Pos': '{0,0,1}',
            },
            1: {
                'Args': "30 -f 3 -b 4 -z",
                'Pos': '{1,0,3}',
            },
            2: {
                'Args': "30 -f a -b z -z",
                'Pos': '{2,0,a}',
            },
        },
    },

    "late2mat2": {
        "args": '30 -f $(foo) -b $(bar) -z $(baz)',

        "job": """
            My.Pos = "{$(Row),$(Step),$(foo)}"
            hold = 1
            max_materialize = 3

            queue foo,bar from (
                1 2
                3 4
                a z
            )
        """,

        "digest": """FACTORY.Requirements=MY.Requirements
args=30 -f $(foo) -b $(bar) -z 
FACTORY.Iwd={IWD}
hold=1
max_materialize=3
My.Pos=\"{{$(Row),$(Step),$(foo)}}\"

Queue 1 foo,bar from """,

        "values": {
            0: {
                'args': '30 -f 1 -b 2 -z',
                'Pos': '{0,0,1}',
            },
            1: {
                'args': '30 -f 3 -b 4 -z',
                'Pos': '{1,0,3}',
            },
            2: {
                'args': '30 -f a -b z -z',
                'Pos': '{2,0,a}',
            },
        },
    },

    "latemat3": {
        "args": None,

        "job": """
            args=30
            My.Pos = "{$(Row),$(Step),$(Item)}"
            hold = 1
            request_memory = $(Item)
            max_materialize = 3

            queue in ( 1, 2, 3 )
        """,

        "digest": """FACTORY.Requirements=MY.Requirements
FACTORY.Iwd={IWD}
hold=1
max_materialize=3
My.Pos=\"{{$(Row),$(Step),$(Item)}}\"
request_memory=$(Item)

Queue 1 Item from """,

        "values": {
            0: {
                'args': '30',
                'Pos': '{0,0,1}',
            },
            1: {
                'args': '30',
                'Pos': '{1,0,2}',
            },
            2: {
                'args': '30',
                'Pos': '{2,0,3}',
            },
        },
    },
}


@action(params={name: name for name in TEST_CASES})
def the_test_case(request):
    return (request.param, TEST_CASES[request.param])


@action
def the_test_name(the_test_case):
    return the_test_case[0]


@action
def the_test_job(the_test_case):
    return the_test_case[1]['job']


@action
def the_test_digest(the_test_case):
    # return the_test_case[1]['digest']
    digest = the_test_case[1]['digest']
    return digest.format(IWD=os.getcwd())

@action
def the_test_values(the_test_case):
    return the_test_case[1]['values']


@action
def the_test_args(the_test_case):
    return the_test_case[1]['args']


@action
def the_submitted_job(path_to_sleep, the_condor, the_test_job, the_test_args):

    submit = htcondor.Submit(
        f"executable = {path_to_sleep}\n" + the_test_job
    )
    # Just to make things harder, apparently.
    if the_test_args is not None:
        submit['args'] = the_test_args

    schedd = the_condor.get_local_schedd()
    return schedd.submit(
        submit,
        # This isn't necessary in version 2.
        count=1,
        # That this is required is actually just broken in version 1.
        itemdata=submit.itemdata(),
    )


@action
def the_expected_digest(the_submitted_job, the_condor, the_test_digest):
    the_itemdata_path = the_condor.spool_dir / str(the_submitted_job.cluster()) / f"condor_submit.{the_submitted_job.cluster()}.items"

    return the_test_digest + the_itemdata_path.as_posix() + "\n"


@action
def the_expected_values(the_test_values):
    return the_test_values


class TestLMMore:

    def test_has_correct_digest(self, the_condor, the_submitted_job, the_expected_digest):
        the_digest_path = the_condor.spool_dir / str(the_submitted_job.cluster()) / f"condor_submit.{the_submitted_job.cluster()}.digest"
        the_digest = the_digest_path.read_text()
        assert the_digest == the_expected_digest


    def test_is_factory_job(self, the_condor, the_submitted_job):
        schedd = the_condor.get_local_schedd()
        constraint = 'ProcID is undefined && JobMaterializeDigestFile isnt undefined'
        results = schedd.query(
            constraint=constraint,
            projection=['ClusterID'],
            opts=htcondor.QueryOpts.IncludeClusterAd,
        )
        clusterIDs = [ad['ClusterID'] for ad in results]
        assert the_submitted_job.cluster() in clusterIDs


    def test_all_jobs_generated_properly(self, the_condor, the_submitted_job, the_expected_values):
        schedd = the_condor.get_local_schedd()
        constraint = f"ClusterID == {the_submitted_job.cluster()}"
        projection = set([ key  for value in the_expected_values.values() for key in value.keys()])
        projection.add('ProcID')
        projection = list(projection)
        results = schedd.query(
            constraint=constraint,
            projection=projection,
        )

        for result in results:
            procID = result['ProcID']
            attributes = the_expected_values[procID]
            for key in attributes.keys():
                assert result[key] == attributes[key]



import sys
import enum


class AdType(enum.IntEnum):
    """
    An enumeration of known ad types.

    .. attribute:: Any

        Matches any type of ad.

    .. attribute:: Generic

    .. attribute:: Slot

        Slot ads, produced by the *condor_startd* daemon.  Used for
        matchmaking.

    .. attribute:: StartDaemon

        Ads about the *condor_startd* daemon itself.  Used for
        location requests and monitoring.

    .. attribute:: Startd

        Either type of ad produced by the *condor_startd* daemon.  Used
        for backwards compatibility.  Use :const:`AdType.Slot` or
        :const:`AdType.StartDaemon` to be explicit.

    .. attribute:: StartdPrivate

    .. attribute:: Schedd

    .. attribute:: Master

    .. attribute:: Collector

    .. attribute:: Negotiator

    .. attribute:: Submitter

    .. attribute:: Grid

    .. attribute:: HAD

    .. attribute:: License

    .. attribute:: Credd

    .. attribute:: Placementd

    .. attribute:: Defrag

    .. attribute:: Accounting

    """

    Startd = 0
    Schedd = 1
    Master = 2
    # Gateway = 3
    # CkptSrvr = 4
    StartdPrivate = 5
    # Submittor = 6
    Submitter = 6
    Collector = 7
    License = 8
    # Storage = 9
    Any = 10
    # bogus = 11
    # Cluster = 12
    Negotiator = 13
    HAD = 14
    Generic = 15
    Credd = 16
    # Database = 17
    # TT = 18
    Grid = 19
    Placementd = 20
    # LeaseManager = 21
    Defrag = 22
    Accounting = 23
    Slot = 24
    StartDaemon = 25

    @staticmethod
    def from_MyType(my_type : str) -> "AdType":
        AdTypeToMyType = {
            AdType.Startd: "Machine",
            AdType.Schedd: "Scheduler",
            AdType.Master: "DaemonMaster",
            AdType.StartdPrivate: "MachinePrivate",
            AdType.Submitter: "Submitter",
            AdType.Collector: "Collector",
            AdType.License: "License",
            AdType.Any: "Any",
            AdType.Negotiator: "Negotiator",
            AdType.HAD: "HAD",
            AdType.Generic: "Generic",
            AdType.Credd: "CredD",
            AdType.Grid: "Grid",
            AdType.Placementd: "PlacementD",
            AdType.Accounting: "Accounting",
            AdType.Slot: "Machine",
            AdType.StartDaemon: "StartD",
        }
        MyTypeToAdType = dict(map(reversed, AdTypeToMyType.items()))

        return MyTypeToAdType.get(my_type)



#!/usr/bin/env pytest


import logging

from ornithology import *

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# Plan for this test is to stand up a condor with 20
# static slots, and configure the NEGOTIATOR_SLOT_CONSTRAINT
# to only look at slot 7 (randomly chosen).  We then
# submit a sleep 0 job, wait for it to complete, leave
# it in the queue, and check to see that it ran on slot7.

@standup
def condor(test_dir):
    raw_config = "USE FEATURE : StaticSlots"
    with Condor(test_dir / "condor", 
            raw_config=raw_config,
            config={
                "NUM_CPUS": "20",
                "NEGOTIATOR_SLOT_CONSTRAINT": "SlotID == 7",
            }) as condor:
        yield condor

@action
def job_one_handle(condor, test_dir, path_to_sleep):
    handle = condor.submit(
        description={
            "log":                      test_dir / "job.log",
            "leave_in_queue":           True,

            "universe":                 "vanilla",
            "should_transfer_files":    False,
            "executable":               path_to_sleep,
            "arguments":                0,
            "Request_Cpus":             1,
            "Request_Memory":           32,
            "Request_Disk":             10240,
        },
        count=1,
    )

    handle.wait(
        condition=ClusterState.all_complete,
        fail_condition=ClusterState.any_held,
        verbose=True,
        timeout=60,
    )

    yield handle

@action
def job_one_ad(job_one_handle):
    for ad in job_one_handle.query():
        return ad

class TestNegotiatorSlotConstraint:
    def test_internal_consistency(self, job_one_ad):
        # There's a race condition when the sched moves
        # RemoteHost to LastRemoteHost. If the latter
        # doesn't exist, try the former
        if "LastRemoteHost" in job_one_ad:
            host = job_one_ad["LastRemoteHost"]
        else:
            host = job_one_ad["RemoteHost"]
        assert host.startswith("slot7@")



#!/usr/bin/env pytest

#
# This test checks condor_ssh_to_job -sftp
# We launch a sleep job, sftp a file to the job
# then from the job.

import htcondor2 as htcondor

from ornithology import (
    config,
    standup,
    action,
    JobStatus,
    ClusterState,
    Condor,
    logging,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

@standup
def condor(test_dir):
    with Condor(test_dir / "condor") as condor:
        yield condor


@action
def sleep_job_parameters(path_to_sleep):
    return {
        "executable": path_to_sleep,
        "transfer_executable": "true",
        "should_transfer_files": "true",
        "universe": "vanilla",
        "arguments": "600",
        "request_cpus": "1",
        "request_disk": "100",
        "request_memory": "1024",
        "log": "sleep.log",
    }

@action
def sleep_job(sleep_job_parameters, condor):
    job = condor.submit( sleep_job_parameters)
    assert job.wait(
        condition=ClusterState.all_running,
        timeout=600,
        verbose=True,
        fail_condition=ClusterState.any_held,
    )
    return job

@action
def ssh_ls(condor, sleep_job):
    cp = condor.run_command(['condor_ssh_to_job', '-auto-retry', '1.0', '/bin/ls'])
    assert cp.returncode == 0
    return True

@action
def sftp_ssh_ls(condor, sleep_job):
    cp = condor.run_command(['/bin/echo', 'ls', '|', 'condor_ssh_to_job', '-ssh', 'sftp', '1.0'])
    assert cp.returncode == 0
    return True

@action
def local_file():
    with open("local_file", "w") as f:
        f.write("This is a local file\n")

@action
def scp_ssh_put(condor, sleep_job, local_file):
    cp = condor.run_command(['condor_ssh_to_job', '-ssh', 'scp', '1.0', 'local_file', 'remote:sandbox_copy'])
    assert cp.returncode == 0
    return True

@action
def scp_ssh_get(condor, sleep_job):
    cp = condor.run_command(['condor_ssh_to_job', '-ssh', 'scp', '1.0', 'remote:sandbox_copy', 'local_copy'])
    assert cp.returncode == 0
    return True

class TestSshSftp:
    def test_ssh(self, ssh_ls):
        assert True 
    def test_sftp_put(self, scp_ssh_put):
        assert True 
    def test_scp_put(self, scp_ssh_put):
        assert True 
    def test_scp_get(self, scp_ssh_get):
        assert True 





#!/usr/bin/env pytest

from ornithology import *
import os
from shutil import rmtree, move
from getpass import getuser

import htcondor2

TOKEN_FILES = ["scitokens.top", "scitokens.use"]
USERNAME = getuser()
TEST_CONFIG = {
    "SEC_CREDENTIAL_DIRECTORY_OAUTH":       "$(LOCAL_DIR)/oauth_credentials",
    "CREDMON_OAUTH_LOG":                    "$(LOG)/CredMonOAuthLog",
    "DAEMON_LIST":                          "$(DAEMON_LIST),CREDMON_OAUTH",
    "AUTO_INCLUDE_CREDD_IN_DAEMON_LIST":    "True",
    "TRUST_CREDENTIAL_DIRECTORY":           "True",
    "LOCAL_CREDMON_PROVIDER_NAME":          "scitokens",
    "LOCAL_CREDMON_TOKEN_AUDIENCE":         "https://localhost",
    "ALLOW_DAEMON":                         "*",
    "LOCAL_CREDMON_PRIVATE_KEY":            "$(LOCAL_DIR)/trust_domain_ca_privkey.pem",
    "LOCAL_CREDMON_PUBLIC_KEY":             "$(LOCAL_DIR)/trust_domain_ca_pubkey.pem",
    "CREDD_DEBUG":                          "D_FULLDEBUG",
    "DAGMAN_MAX_SUBMIT_ATTEMPTS":           "1",
    "DAGMAN_PRODUCE_JOB_CREDENTIALS":       "True",
}

@standup
def condor(test_dir):
    with Condor(local_dir=test_dir / "condor", config=TEST_CONFIG) as condor:
        yield condor

def write_submit(exe: str):
    filename = "sleep.sub"
    with open(filename, "w") as f:
        f.write(
        f"""
        executable = {exe}
        arguments  = 1
        log        = job.log
        queue
        """)
    return filename

def write_dag(submit: str):
    filename = "test.dag"
    with open(filename, "w") as f:
        f.write(f"JOB ONLY {submit}\n")
    return filename

TEST_CASES = {
    "shell_condor_submit" : {
        "DAG-Options" : {"SubmitMethod" : 0},
    },
    "direct_submit" : {
        "DAG-Options" : {"SubmitMethod" : 1},
    },
}

@action(params={name: name for name in TEST_CASES})
def run_dag(condor, path_to_sleep, test_dir, request):
    os.chdir(str(test_dir))
    sub_dir = os.path.join(str(test_dir), request.param)
    if os.path.exists(sub_dir):
        rmtree(sub_dir)
    os.mkdir(sub_dir)
    os.chdir(sub_dir)

    submit_file = write_submit(str(path_to_sleep))
    dag_file = write_dag(submit_file)

    with condor.use_config():
        creds_dir = htcondor2.param["SEC_CREDENTIAL_DIRECTORY_OAUTH"]
        DAG = htcondor2.Submit.from_dag(dag_file, options=TEST_CASES[request.param]["DAG-Options"])

    for token_file in TOKEN_FILES:
        path = os.path.join(creds_dir, USERNAME, token_file)
        assert not os.path.exists(path)

    dag_handle = condor.submit(DAG)
    assert dag_handle.wait(condition=ClusterState.all_complete, timeout=60)

    for token_file in TOKEN_FILES:
        path = os.path.join(creds_dir, USERNAME, token_file)
        try:
            move(path, ".")
        except:
            pass

    return sub_dir


class TestDAGManCredentials:
    def test_dagman_produce_creds(self, run_dag):
        dir_contents = os.listdir(run_dag)
        for token_file in TOKEN_FILES:
            assert token_file in dir_contents



import pickle
import threading
from urllib.error import HTTPError
from urllib.request import urlopen
import json
from filelock import FileLock
import hashlib
import time
import os
from flask import request, make_response, Response
import datetime
from wsgiref.handlers import format_date_time

app = None  # Placeholder for the Flask app instance

def set_flask_app(flask_app):
    global app
    app = flask_app

# Implementation of method decorator on a flask response function
def cache_response_to_disk(seconds_to_cache: int = None, file_name: str = None):
    """
    A decorator to cache the response of a function to disk for a specified duration.
    Args:
        seconds_to_cache (int, optional): The number of seconds to cache the response. 
            If not provided, defaults to the value of `CE_DASHBOARD_SERVER_CACHE_MINUTES` 
            from the application configuration multiplied by 60.
        file_name (str, optional): The name of the cache file. If not provided, a unique 
            cache file name is generated based on the query string.
    Returns:
        function: A wrapper function that caches the response of the decorated function 
        to a file on disk.
    Notes:
        - The cache directory is determined by the `CE_DASHBOARD_CACHE_DIRECTORY` 
          configuration in the application.
        - If the cache file does not exist or is expired, the decorated function is 
          executed, and its response is cached.
        - If the cache file exists and is valid, the cached response is returned.
        - A file lock is used to ensure thread-safe access to the cache file.
        - If an error occurs while writing to the cache file (e.g., disk full), the 
          cache file is removed.
    Returns:
        tuple: A tuple containing:
            - response (str): The response from the decorated function or the cached response.
            - cached_response (bool): A flag indicating whether the response was retrieved 
              from the cache (True) or generated by the function (False).
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            CACHE_DIR = app.config['CE_DASHBOARD_CACHE_DIRECTORY']
            effective_seconds_to_cache = seconds_to_cache
            if effective_seconds_to_cache is None:
                effective_seconds_to_cache = app.config['CE_DASHBOARD_SERVER_CACHE_MINUTES'] * 60
            if file_name is not None:
                cache_file = os.path.join(CACHE_DIR, file_name)
            else:
                # Create a unique cache file name based on the query string
                query_string = request.query_string.decode('utf-8')
                query_hash = hashlib.md5(query_string.encode()).hexdigest()
                cache_file = os.path.join(CACHE_DIR, query_hash)

            cached_response = True

            # Create cache directory if it does not exist
            os.makedirs(os.path.dirname(cache_file), exist_ok=True)

            # if filename ends with '.bin', use binary mode
            open_mode = 'a+'
            if cache_file.endswith('.bin'):
                open_mode = 'ab+'

            lock = FileLock(cache_file + ".lock",is_singleton=False)
            with lock:
                with open(cache_file, open_mode) as f:
                    fsize = os.path.getsize(cache_file)
                    fmtime = os.path.getmtime(cache_file)
                    if (fsize == 0) or \
                       (fmtime < (time.time() - effective_seconds_to_cache)):
                        cached_response = False
                        response = func(*args, **kwargs)
                        try:
                            f.seek(0)
                            f.truncate()
                            f.write(response)
                        except Exception as e:
                            # Remove cache file if any errors writing it (i.e. disk full)
                            os.remove(cache_file)
                            print(f"ERROR writing cache file {cache_file}: {e}")
                    else:
                        f.seek(0)
                        response = f.read()
            return response, cached_response
        return wrapper
    return decorator


def make_data_response(response_body: str, cached_response: bool, browser_cache_minutes: int = None) -> Response:
    """
    Creates an HTTP response for with specified caching headers and metadata.

    Args:
        response_body (str): The body of the HTTP response.
        cached_response (bool): Indicates whether the response is from the server cache.
        browser_cache_minutes (int, optional): The number of minutes the browser should cache the response.

    Returns:
        Response: A Flask Response object with the specified headers and content.

    Notes:
        - The 'Content-Type' header is set to "text/plain; charset=UTF-8".
        - The 'Expires' header specifies when the cached response should expire.
        - The 'Cache-Control' header defines the maximum age for browser caching.
        - A custom header 'X-HTCondorView-Cached-Response' indicates if the response is cached.
    """
    if (browser_cache_minutes is None):
        browser_cache_minutes = app.config['CE_DASHBOARD_BROWSER_CACHE_MINUTES']
    response = make_response(response_body)
    response.headers['Content-Type'] = "text/plain; charset=UTF-8"
    # Tell the browser to cache this data for a specified number of minutes
    browser_cache_seconds = browser_cache_minutes * 60
    now = datetime.datetime.now()
    expires = now + datetime.timedelta(minutes=browser_cache_minutes)
    expires_str = format_date_time(time.mktime(expires.timetuple()))
    response.headers['Expires'] = f"{expires_str}"
    response.headers['Cache-Control'] =  f"max-age={browser_cache_seconds}, public"
    # In a custom HTTP header, state if the response is from the cache or not
    response.headers['X-HTCondorView-Cached-Response'] = f"{cached_response}"
    return response


institution_db_lock = threading.Lock()
institution_db = {}

@cache_response_to_disk(file_name="institution_db.bin", seconds_to_cache=60*60)
def _getInstitutionDBPickle() -> bytes:
    tries = 0
    max_tries = 2
    INSTITUTION_DATABASE_URL = "https://topology-institutions.osg-htc.org/api/institution_ids"
    while tries < max_tries:
        try:
            with urlopen(INSTITUTION_DATABASE_URL) as f:
                for institution in json.load(f):
                    institution_id = institution.get("id")
                    if not institution_id:
                        continue
                    institution_id_short = institution_id.split("/")[-1]
                    institution["id_short"] = institution_id_short
                    institution_db[institution_id] = institution
                    institution_db[institution_id_short] = institution

                    # OSG_INSTITUTION_IDS mistakenly had the ROR IDs before ~2024-11-07,
                    # so we map those too (as long as they don't conflict with OSG IDs)
                    ror_id_short = (institution.get("ror_id") or "").split("/")[-1]
                    if ror_id_short and ror_id_short not in institution_db:
                        institution_db[ror_id_short] = institution
                print("INFO: Loaded institution database with %d entries" % len(institution_db))
        except HTTPError:
            time.sleep(2**tries)
            tries += 1
            if tries == max_tries and len(institution_db) == 0:
                raise
        else:
            break

    # Pickle institution_db into a binary string
    binary_data = pickle.dumps(institution_db)
    return binary_data

def getOrganizationFromInstitutionID(institution_id: str, default: str) -> str:
    """
    Returns the organization name from the institution ID.

    Args:
        institution_id (str): The institution ID.

    Returns:
        str: The organization name.
    """

    if institution_id == "Unknown":
        return default

    with institution_db_lock:
        global institution_db
        if institution_id in institution_db:
            return institution_db[institution_id].get("name", default)
        else:
            # If the institution ID is not found, reload the database
            # and try again. This is a fallback mechanism to ensure that the
            # institution database is up to date.
            institution_db = {}
            buffer, _ = _getInstitutionDBPickle()
            institution_db.update(pickle.loads(buffer))
            if institution_id in institution_db:
                return institution_db[institution_id].get("name", default)
            else:
                # If still not found, return the default value
                # and log a warning message.
                print(f"WARNING: Institution ID {institution_id} not found in database.")   

    return default



import sys
import enum


class VacateType(enum.IntEnum):
    """
    An enumeration of vacate policies that can be sent to a *condor_startd*.

    The following explanations are reminders; the
    `actual policies <http://htcondor.readthedocs.io/en/latest/admin-manual/ep-policy-configuration.html#machine-states>`_
    are way more complicated.

    .. attribute:: Graceful

        Jobs are immediately soft-killed (generally meaning sent a ``SIGTERM``).
        The startd then waits a certain amount of time to allow jobs to shut
        down cleanly before the vacate policy becomes the following.

    .. attribute:: Fast

        Jobs are immediately hard-killed (generally meaning sent a ``SIGKILL``).
    """

    Graceful = 0
    Fast     = 1



import os
import sys

from docutils import nodes
from docutils.parsers.rst import Directive
from sphinx import addnodes
from sphinx.errors import SphinxError
from sphinx.util.nodes import split_explicit_title, process_index_entry, set_role_source_info
from htc_helpers import *

CLASSAD_FUNCTIONS = []

def find_classad_func_cmds(dir: str):
    classad_functions = []
    classad_ref = os.path.join(dir, "classads", "classad-builtin-functions.rst")
    with open(classad_ref, "r", encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            for cmd in get_all_defined_role("classad-function-def", line):
                if "(" not in cmd or " " not in cmd[:cmd.find("(")]:
                    continue
                end = cmd.find("(")
                begin = cmd[:end].find(" ") + 1
                function = cmd[begin : end]
                if function not in classad_functions:
                    classad_functions.append(function)
    classad_functions.sort()
    return classad_functions

def dump(obj):
    for attr in dir(obj):
        print("obj.%s = %r" % (attr, getattr(obj, attr)))

def classad_function_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    root_dir = root_dir = get_rel_path_to_root_dir(inliner)[:-1]
    function, index = custom_ext_parser(text)

    # Allow :classad-function:`foo` and :classad-function:`foo()`
    if "(" in function:
        function = function[:function.find("(")]

    if function not in CLASSAD_FUNCTIONS:
        docname = inliner.document.settings.env.docname
        warn(f"{docname}:{lineno} | '{function}' ClassAd function not in defined list. Either a typo or not defined.")

    ref_link = f"href=\"{root_dir}/classads/classad-builtin-functions.html#" + str(function) + "()\""
    return make_ref_and_index_nodes(name, function, index, ref_link,
                                    rawtext, inliner, lineno, options, f"{function}()")

def setup(app):
    global CLASSAD_FUNCTIONS
    CLASSAD_FUNCTIONS = find_classad_func_cmds(app.srcdir)
    app.add_role("classad-function", classad_function_role)




#!/usr/bin/env pytest

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

from ornithology import (
    Condor,
    action,
    ClusterState,
)

import time
import htcondor2
import classad2


@action
def the_lock_file(test_dir):
	return test_dir / 'job.started'


@action
def the_condor(test_dir):
	local_dir = test_dir / "condor"
	with Condor(
		local_dir=local_dir,
		config={
			'STARTD_DEBUG':					'D_CATEGORY D_SUBSECOND D_TEST',
			'STARTER_DEBUG':				'D_CATEGORY D_SUBSECOND D_TEST',
			'TESTING_COLOR_FROM_JOB_AD':	'TRUE',
			# By making this explicit, I don't have to look it up later.
			'UPDATE_INTERVAL':				'2',
		},
	) as the_condor:
		yield the_condor


@action
def the_running_job(test_dir, the_lock_file, the_condor):
	job_description = {
		'shell':						f'touch {the_lock_file.as_posix()}; sleep 300',
		'should_transfer_files':		'YES',
		'log':							test_dir / 'job.log',
		'MY.ColorAd':					'[ ColorAttr = 7 ]',
	}

	job_handle = the_condor.submit(
		description=job_description,
		count=1,
	);

	# Wait for the job to be marked running...
	assert job_handle.wait(
		timeout=20,
		condition=ClusterState.all_running,
	)

	# ... wait for the job to have actually started.
	for i in range(0,20):
		if the_lock_file.exists():
			break
		time.sleep(1)
	assert the_lock_file.exists()

	return job_handle


@action
def the_slot_ads(the_condor, the_running_job):
	# Poll until the collector has slot ads with ColorAttr set.
	for i in range(0, 30):
		result = the_condor.status(
			ad_type=htcondor2.AdType.Slot,
		)
		if len(result) > 0 and all(ad.get('ColorAttr', 0) == 7 for ad in result):
			break
		time.sleep(1)

	assert(len(result) > 0)

	return result


@action
def the_extra_ads(the_condor, the_running_job):
	result = the_condor.run_command(
		args=['condor_who', '-l', '-snapshot'],
	)
	assert result.returncode == 0

	ads = list(classad2.parseAds(result.stdout, classad2.ParserType.Old))
	assert(len(ads) > 0)

	ads = [ad for ad in ads if ad.get('MyType', "") == "Machine.Extra"]
	assert(len(ads) > 0)

	return ads


@action
def the_cleaned_up_ads(
	the_condor, the_running_job,
	# Make sure that we've recorded these before we perturb the state.
	the_extra_ads, the_slot_ads
):
	result_ad = the_running_job.remove()
	# assert something about the result_ad here

	# Poll until the collector has slot ads with ColorAttr removed.
	for i in range(0, 30):
		result = the_condor.status(
			ad_type=htcondor2.AdType.Slot,
		)
		if len(result) > 0 and all(ad.get('ColorAttr', -1) == -1 for ad in result):
			break
		time.sleep(1)

	assert(len(result) > 0)

	return result


class TestSlotColoring:

	# This is an implementation details, but may be useful for debugging.
	def test_extra_ads(self, the_extra_ads):
		for extra_ad in the_extra_ads:
			assert extra_ad.get('ColorAttr', 0) == 7


	def test_slot_ads(self, the_slot_ads):
		for slot_ad in the_slot_ads:
			assert slot_ad.get('ColorAttr', 0) == 7


	def test_color_cleanup(self, the_cleaned_up_ads):
		for slot_ad in the_cleaned_up_ads:
			assert slot_ad.get('ColorAttr', -1) == -1



import sys
import enum


class LogLevel(enum.IntEnum):
    """
    An enumeration of "log levels" to use with :func:`log`.  Some
    values of this enumeration select types of log messages; others
    control the format of the header.  See
    `the documentation <https://htcondor.readthedocs.io/en/latest/admin-manual/configuration-macros.html#%3CSUBSYS%3E_DEBUG>`_.

    .. attribute:: Always
    .. attribute:: Audit
    .. attribute:: Config
    .. attribute:: DaemonCore
    .. attribute:: Error
    .. attribute:: FullDebug
    .. attribute:: Hostname
    .. attribute:: Job
    .. attribute:: Machine
    .. attribute:: Network
    .. attribute:: NoHeader
    .. attribute:: PID
    .. attribute:: Priv
    .. attribute:: Protocol
    .. attribute:: Security
    .. attribute:: Status
    .. attribute:: SubSecond
    .. attribute:: Terse
    .. attribute:: Timestamp
    .. attribute:: Verbose
    """

    Always = 0
    Error = 1
    Status = 2
    # ZKM = 3
    Job = 4
    Machine = 5
    Config = 6
    Protocol = 7
    Priv = 8
    DaemonCore = 9
    Verbose = 10
    Security = 11
    # Command = 12
    # Match = 13
    # If D_COMMAND and D_MATCH aren't on the list, why is D_NETWORK?
    Network = 14
    # Keyboard = 15
    # ProcFamily = 16
    # Idle = 17
    # Threads = 18
    # Accountant = 19
    # Syscalls = 20
    # Reserved1 = 21
    Hostname = 22
    # Why is D_PERF_TRACE missing?
    # PerfTrace = 23
    Audit = 24
    # Test = 25
    # Stats = 26
    # Materialize = 27
    # Bug = 28

    Terse = (0<<8)
    # Verbose = (1<<8)
    # Diagnostic = (2<<8)
    # Never = (3<<8)

    FullDebug = (1<<10)
    # ErrorAlso = (1<<11)
    # Except = (1<<12)
    # Backtrace = (1<<24)
    # Ident = (1<<25)
    SubSecond = (1<<26)
    Timestamp = (1<<27)
    PID = (1<<28)
    # FDs = (1<<29)
    # CAT = (1<<30)
    NoHeader = (1<<31)



import os

from docutils.parsers.rst import Directive
from docutils.parsers.rst import directives
from docutils import nodes

class FakeDefinition(Directive):
    """Custom directive to be used as filler for multi-term definitions"""
    optional_arguments = 0
    has_content = False

    def run(self) -> list:
        node = nodes.inline("", "", classes=['faux-definition'])
        return [node]

def setup(app):
    """Setup sphinx to know of this directive"""
    app.add_directive('faux-definition', FakeDefinition)
    app.add_directive('fake-definition', FakeDefinition)
